from . import _ffi as ffi
import ctypes
import typing
from wasmtime import MemoryType, WasmtimeError
from ._store import Storelike


class Memory:
    _memory: ffi.wasmtime_memory_t

    def __init__(self, store: Storelike, ty: MemoryType):
        """
        Creates a new memory in `store` with the given `ty`
        """

        mem = ffi.wasmtime_memory_t()
        error = ffi.wasmtime_memory_new(store._context(), ty.ptr(), ctypes.byref(mem))
        if error:
            raise WasmtimeError._from_ptr(error)
        self._memory = mem

    @classmethod
    def _from_raw(cls, mem: ffi.wasmtime_memory_t) -> "Memory":
        ty: "Memory" = cls.__new__(cls)
        ty._memory = mem
        return ty

    def type(self, store: Storelike) -> MemoryType:
        """
        Gets the type of this memory as a `MemoryType`
        """

        ptr = ffi.wasmtime_memory_type(store._context(), ctypes.byref(self._memory))
        return MemoryType._from_ptr(ptr, None)

    def grow(self, store: Storelike, delta: int) -> int:
        """
        Grows this memory by the given number of pages
        """

        if delta < 0:
            raise WasmtimeError("cannot grow by negative amount")
        prev = ctypes.c_uint64(0)
        error = ffi.wasmtime_memory_grow(store._context(), ctypes.byref(self._memory), delta, ctypes.byref(prev))
        if error:
            raise WasmtimeError._from_ptr(error)
        return prev.value

    def size(self, store: Storelike) -> int:
        """
        Returns the size, in WebAssembly pages, of this memory.
        """

        return ffi.wasmtime_memory_size(store._context(), ctypes.byref(self._memory))

    def data_ptr(self, store: Storelike) -> "ctypes._Pointer[ctypes.c_ubyte]":
        """
        Returns the raw pointer in memory where this wasm memory lives.

        Remember that all accesses to wasm memory should be bounds-checked
        against the `data_len` method.
        """
        return ffi.wasmtime_memory_data(store._context(), ctypes.byref(self._memory))

    def get_buffer_ptr(self, store: Storelike,
                       size: typing.Optional[int] = None,
                       offset: int = 0) -> ctypes.Array:
        """
        return raw pointer to buffer suitable for creating zero-copy writable NumPy Buffer Protocol
        this method is also used internally by `read()` and `write()`

        np_mem = np.frombuffer(memory.get_buffer_ptr(store), dtype=np.uint8)
        np_mem[start:end] = A # write
        B = np_mem[start:end] # read
        """
        if size is None:
            size = self.data_len(store)
        ptr_type = ctypes.c_ubyte * size
        return ptr_type.from_address(ctypes.addressof(self.data_ptr(store).contents) + offset)

    def read(
            self,
            store: Storelike,
            start: typing.Optional[int] = 0,
            stop: typing.Optional[int] = None) -> bytearray:
        """
        Reads this memory starting from `start` and up to `stop`
        and returns a copy of the contents as a `bytearray`.

        The indexing behavior of this method is similar to `list[start:stop]`
        where negative starts can be used to read from the end, for example.
        """
        size = self.data_len(store)
        key = slice(start, stop, None)
        start, stop, _ = key.indices(size)
        val_size = stop - start
        if val_size <= 0:
            # return bytearray of size zero
            return bytearray(0)
        src_ptr = self.get_buffer_ptr(store, val_size, start)
        return bytearray(src_ptr)

    def write(
            self,
            store: Storelike,
            value: typing.Union[bytearray, bytes],
            start: typing.Optional[int] = None) -> int:
        """
        write a bytearray value into a possibly large slice of memory
        negative start is allowed in a way similat to list slice mylist[-10:]
        if value is not bytearray it will be used to construct an intermediate bytearray (copyied twice)
        return number of bytes written
        raises IndexError when trying to write outside the memory range
        this happens when start offset is >= size or when end side of value is >= size
        """
        size = self.data_len(store)
        key = slice(start, None)
        start = key.indices(size)[0]
        if start >= size:
            raise IndexError("index out of range")
        # value must be bytearray ex. cast bytes() to bytearray
        if not isinstance(value, bytearray):
            value = bytearray(value)
        val_size = len(value)
        if val_size == 0:
            return val_size
        # stop is exclusive
        stop = start + val_size
        if stop > size:
            raise IndexError("index out of range")
        ptr_type = ctypes.c_ubyte * val_size
        src_ptr = ptr_type.from_buffer(value)
        dst_ptr = self.get_buffer_ptr(store, val_size, start)
        ctypes.memmove(dst_ptr, src_ptr, val_size)
        return val_size

    def data_len(self, store: Storelike) -> int:
        """
        Returns the raw byte length of this memory.
        """

        return ffi.wasmtime_memory_data_size(store._context(), ctypes.byref(self._memory))

    def _as_extern(self) -> ffi.wasmtime_extern_t:
        union = ffi.wasmtime_extern_union(memory=self._memory)
        return ffi.wasmtime_extern_t(ffi.WASMTIME_EXTERN_MEMORY, union)



from .. import _ffi as ffi
from .._wat2wasm import _to_wasm
from ._types import ComponentType
import ctypes
from wasmtime import Engine, wat2wasm, WasmtimeError, Managed, Module
import typing
from os import PathLike


class ExportIndex(Managed["ctypes._Pointer[ffi.wasmtime_component_export_index_t]"]):

    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct an `ExportIndex`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_export_index_t]") -> None:
        ffi.wasmtime_component_export_index_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_export_index_t]") -> "ExportIndex":
        if not isinstance(ptr, ctypes.POINTER(ffi.wasmtime_component_export_index_t)):
            raise TypeError("wrong pointer type")
        ty: "ExportIndex" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty


class Component(Managed["ctypes._Pointer[ffi.wasmtime_component_t]"]):

    @classmethod
    def from_file(cls, engine: Engine, path: typing.Union[str, bytes, PathLike]) -> "Component":
        """
        Compiles and creates a new `Component` by reading the file at `path` and
        then delegating to the `Component` constructor.
        """

        with open(path, "rb") as f:
            contents = f.read()
        return cls(engine, contents)

    def __init__(self, engine: Engine, wasm: typing.Union[str, bytes, bytearray]):
        if not isinstance(engine, Engine):
            raise TypeError("expected an Engine")

        wasm = _to_wasm(wasm)

        # TODO: can the copy be avoided here? I can't for the life of me
        # figure this out.
        binary = (ctypes.c_uint8 * len(wasm)).from_buffer_copy(wasm)
        ptr = ctypes.POINTER(ffi.wasmtime_component_t)()
        error = ffi.wasmtime_component_new(engine.ptr(), binary, len(wasm), ctypes.byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        self._set_ptr(ptr)

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_t]") -> None:
        ffi.wasmtime_component_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_t]") -> "Component":
        if not isinstance(ptr, ctypes.POINTER(ffi.wasmtime_component_t)):
            raise TypeError("wrong pointer type")
        ty: "Component" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @classmethod
    def deserialize(cls, engine: Engine, encoded: typing.Union[bytes, bytearray]) -> 'Component':
        """
        Deserializes bytes previously created by `Component.serialize`.

        This constructor for `Component` will deserialize bytes previously created
        by a serialized component. This will only succeed if the bytes were
        previously created by the same version of `wasmtime` as well as the
        same configuration within `Engine`.
        """

        if not isinstance(encoded, (bytes, bytearray)):
            raise TypeError("expected bytes")

        ptr = ctypes.POINTER(ffi.wasmtime_component_t)()

        # TODO: can the copy be avoided here? I can't for the life of me
        # figure this out.
        error = ffi.wasmtime_component_deserialize(
            engine.ptr(),
            (ctypes.c_uint8 * len(encoded)).from_buffer_copy(encoded),
            len(encoded),
            ctypes.byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        return cls._from_ptr(ptr)

    @classmethod
    def deserialize_file(cls, engine: Engine, path: str) -> 'Component':
        """
        Deserializes bytes previously created by `Component.serialize` that are
        stored in a file on the filesystem.

        Otherwise this function is the same as `Component.deserialize`.
        """

        ptr = ctypes.POINTER(ffi.wasmtime_component_t)()
        path_bytes = path.encode('utf-8')
        error = ffi.wasmtime_component_deserialize_file(
            engine.ptr(),
            path_bytes,
            ctypes.byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        return cls._from_ptr(ptr)

    def serialize(self) -> bytearray:
        """
        Serializes this component to a binary representation.

        This method will serialize this component to an in-memory byte array
        which can be cached and later passed to `Component.deserialize` to
        recreate this component.
        """
        raw = ffi.wasm_byte_vec_t()
        err = ffi.wasmtime_component_serialize(self.ptr(), ctypes.byref(raw))
        if err:
            raise WasmtimeError._from_ptr(err)
        ret = ffi.to_bytes(raw)
        ffi.wasm_byte_vec_delete(ctypes.byref(raw))
        return ret

    def get_export_index(self, name: str, instance : typing.Optional[ExportIndex] = None) -> typing.Optional[ExportIndex]:
        """
        Gets an `ExportIndex` from this component pointing to a specific item
        in this component.

        The returned `ExportIndex` can later be used to lookup exports on an
        instance or it can be used to lookup further indexes if it points to an
        instance for example.
        """
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        ret = ffi.wasmtime_component_get_export_index(
            self.ptr(),
            instance.ptr() if instance is not None else None,
            name_buf,
            len(name_bytes))
        if not ret:
            return None
        return ExportIndex._from_ptr(ret)

    @property
    def type(self) -> ComponentType:
        """
        Returns the `ComponentType` corresponding to this component.
        """
        ptr = ffi.wasmtime_component_type(self.ptr())
        return ComponentType._from_ptr(ptr)



# Example of instantiating two modules which link to each other.

from wasmtime import Engine, Store, Module, Linker, WasiConfig

def run():
    engine = Engine()

    # Load and compile our two modules
    linking1 = Module.from_file(engine, "examples/linking1.wat")
    linking2 = Module.from_file(engine, "examples/linking2.wat")

    # Set up our linker which is going to be linking modules together. We
    # want our linker to have wasi available, so we set that up here as well.
    linker = Linker(engine)
    linker.define_wasi()

    # Create a `Store` to hold instances, and configure wasi state
    store = Store(engine)
    wasi = WasiConfig()
    wasi.inherit_stdout()
    store.set_wasi(wasi)

    # Instantiate our first module which only uses WASI, then register that
    # instance with the linker since the next linking will use it.
    linking2 = linker.instantiate(store, linking2)
    linker.define_instance(store, "linking2", linking2)

    # And with that we can perform the final link and the execute the module.
    linking1 = linker.instantiate(store, linking1)
    run = linking1.exports(store)["run"]
    run(store)

if __name__ == '__main__':
    run()



import ctypes
from ctypes import POINTER, byref, create_string_buffer, CFUNCTYPE, c_void_p
from .. import Managed, Engine, WasmtimeError, Module, _ffi as ffi, Storelike, StoreContext
from .._config import setter_property
from ._instance import Instance
from ._component import Component
from ._resource_type import ResourceType
from typing import Union, Tuple, Callable, Optional, List, Any
from .._slab import Slab
from ._enter import catch_exceptions
from ._types import ValType, FuncType

LinkerInstanceParent = Union['Linker', 'LinkerInstance']

ResourceDtor = Callable[[StoreContext, int], None]
RESOURCE_DTORS: Slab[ResourceDtor] = Slab()

UserFunc = Callable[..., Optional[Any]]
FUNCTIONS: Slab[UserFunc] = Slab()


class Linker(Managed["ctypes._Pointer[ffi.wasmtime_component_linker_t]"]):
    engine: Engine
    locked: bool

    def __init__(self, engine: Engine):
        """
        Creates a new linker ready to instantiate modules within the store
        provided.
        """
        self._set_ptr(ffi.wasmtime_component_linker_new(engine.ptr()))
        self.engine = engine
        self.locked = False

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_linker_t]") -> None:
        self._assert_not_locked()
        ffi.wasmtime_component_linker_delete(ptr)

    @setter_property
    def allow_shadowing(self, allow: bool) -> None:
        """
        Configures whether definitions are allowed to shadow one another within
        this linker
        """
        self._assert_not_locked()
        ffi.wasmtime_component_linker_allow_shadowing(self.ptr(), allow)

    def _assert_not_locked(self) -> None:
        if self.locked:
            raise WasmtimeError("cannot use linker while it's in use by other instances")

    def root(self) -> "LinkerInstance":
        """
        Returns the root instance used to defined new items in this linker.

        While this root instance is alive this linker cannot be used for any
        other operations. Until the returned object is destroyed this linker is
        considered "locked".

        It's recommended to bind the return value in a `with` block to
        automatically dispose of it when it's done.
        """
        self._assert_not_locked()
        ptr = ffi.wasmtime_component_linker_root(self.ptr())
        return LinkerInstance._from_ptr(ptr, self)

    def add_wasip2(self) -> None:
        """
        Adds the WASIp2 API definitions, from Wasmtime, in this linker.
        """
        self._assert_not_locked()
        err = ffi.wasmtime_component_linker_add_wasip2(self.ptr())
        if err:
            raise WasmtimeError._from_ptr(err)

    def instantiate(self, store: Storelike, component: Component) -> Instance:
        """
        Instantiates the given component using this linker within the provided
        store.

        Returns the instantiated `Instance` on success.
        """
        self._assert_not_locked()
        instance = ffi.wasmtime_component_instance_t()
        err = ffi.wasmtime_component_linker_instantiate(
            self.ptr(),
            store._context(),
            component.ptr(),
            byref(instance))
        if err:
            raise WasmtimeError._from_ptr(err)
        return Instance._from_raw(instance)

    def define_unknown_imports_as_traps(self, component: Component) -> None:
        """
        Configures this linker to define any unknown imports of `component` as
        traps which will error when invoked.
        """
        self._assert_not_locked()
        err = ffi.wasmtime_component_linker_define_unknown_imports_as_traps(
            self.ptr(),
            component.ptr())
        if err:
            raise WasmtimeError._from_ptr(err)


class LinkerInstance(Managed["ctypes._Pointer[ffi.wasmtime_component_linker_instance_t]"]):
    parent: Union[LinkerInstanceParent, None]
    locked: bool

    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `LinkerInstance`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_linker_instance_t]") -> None:
        self._assert_not_locked()
        ffi.wasmtime_component_linker_instance_delete(ptr)
        assert(self.parent is not None)
        assert(self.parent.locked)
        self.parent.locked = False
        self.parent = None

    def _assert_not_locked(self) -> None:
        if self.locked:
            raise WasmtimeError("cannot use linker instance while it's in use by other instances")

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_linker_instance_t]", parent: LinkerInstanceParent) -> "LinkerInstance":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_linker_instance_t)):
            raise TypeError("wrong pointer type")
        ty: "LinkerInstance" = cls.__new__(cls)
        ty._set_ptr(ptr)
        ty.parent = parent
        ty.locked = False
        assert(not parent.locked)
        parent.locked = True
        return ty


    def add_instance(self, name: str) -> "LinkerInstance":
        """
        Adds a new instance to this linker instance under the given name.

        Returns a new `LinkerInstance` which can be used to define further
        items.

        While the returned instance is alive this linker instance cannot be
        used for other operations. Until the returned object is destroyed this
        linker instance is considered "locked".

        It's recommended to bind the return value in a `with` block to
        automatically dispose of it when it's done.
        """
        self._assert_not_locked()
        name_bytes = name.encode('utf-8')
        name_buf = create_string_buffer(name_bytes)
        ptr = POINTER(ffi.wasmtime_component_linker_instance_t)()
        err = ffi.wasmtime_component_linker_instance_add_instance(self.ptr(),
                                                                  name_buf,
                                                                  len(name_bytes),
                                                                  byref(ptr))
        if err:
            raise WasmtimeError._from_ptr(err)
        return LinkerInstance._from_ptr(ptr, self)

    def add_module(self, name: str, module: Module) -> None:
        """
        Adds a new module to this linker instance under the given name.
        """
        self._assert_not_locked()
        name_bytes = name.encode('utf-8')
        name_buf = create_string_buffer(name_bytes)
        err = ffi.wasmtime_component_linker_instance_add_module(self.ptr(),
                                                                name_buf,
                                                                len(name_bytes),
                                                                module.ptr())
        if err:
            raise WasmtimeError._from_ptr(err)

    def add_resource(self, name: str, ty: ResourceType, dtor: ResourceDtor) -> None:
        """
        Defines a the resource type `ty` under `name`

        When a value of this resource is destroyed then `dtor` is invoked.
        """
        self._assert_not_locked()
        name_bytes = name.encode('utf-8')
        name_buf = create_string_buffer(name_bytes)
        idx = RESOURCE_DTORS.allocate(dtor)
        err = ffi.wasmtime_component_linker_instance_add_resource(self.ptr(),
                                                                  name_buf,
                                                                  len(name_bytes),
                                                                  ty.ptr(),
                                                                  resource_dtor_impl,
                                                                  idx,
                                                                  resource_dtor_finalize)
        if err:
            raise WasmtimeError._from_ptr(err)

    def add_func(self, name: str, func: UserFunc) -> None:
        self._assert_not_locked()
        name_bytes = name.encode('utf-8')
        name_buf = create_string_buffer(name_bytes)
        idx = FUNCTIONS.allocate(func)
        err = ffi.wasmtime_component_linker_instance_add_func(self.ptr(),
                                                              name_buf,
                                                              len(name_bytes),
                                                              func_impl,
                                                              idx,
                                                              func_finalize)
        if err:
            raise WasmtimeError._from_ptr(err)


@ffi.wasmtime_component_resource_destructor_t
def resource_dtor_impl(idx, context, rep): # type: ignore
    def run(store: StoreContext) -> None:
        return RESOURCE_DTORS.get(idx or 0)(store, rep)
    return catch_exceptions(context, run)


@CFUNCTYPE(None, c_void_p)
def resource_dtor_finalize(idx): # type: ignore
    if RESOURCE_DTORS:
        RESOURCE_DTORS.deallocate(idx or 0)
    return None


@ffi.wasmtime_component_func_callback_t
def func_impl(idx, context, ty_raw, args, nargs, results, nresults): # type: ignore
    def run(store: StoreContext) -> None:
        with FuncType._from_ptr(ty_raw, owner=store) as fty:
            func = FUNCTIONS.get(idx or 0)
            param_tys = fty.params
            result_ty = fty.result
            assert(len(param_tys) == nargs)
            if result_ty is None:
                assert(nresults == 0)
            else:
                assert(nresults == 1)

            pyargs = []
            for (_, ty), i in zip(param_tys, range(nargs)):
                val = ty.convert_from_c(args[i])
                pyargs.append(val)
            result = func(store, *pyargs)

            if result_ty is None:
                assert(result is None)
            else:
                result_ty.convert_to_c(store, result, results)
    return catch_exceptions(context, run)


@CFUNCTYPE(None, c_void_p)
def func_finalize(idx): # type: ignore
    if FUNCTIONS:
        FUNCTIONS.deallocate(idx or 0)
    return None



from ._component import Component, ExportIndex
from ._linker import Linker, LinkerInstance
from ._instance import Instance
from ._func import Func
from ._types import ComponentType, ModuleType, ComponentItem, ComponentInstanceType, FuncType, ResourceType
from ._types import ValType, Bool, U8, U16, U32, U64, S8, S16, S32, S64, F32, F64, Char, String, ErrorContext
from ._types import ListType, RecordType, TupleType, VariantType, EnumType, OptionType, ResultType, FlagsType
from ._types import StreamType, FutureType, OwnType, BorrowType, Variant, Record
from ._resources import ResourceAny, ResourceHost

__all__ = [
    'Component',
    'ExportIndex',
    'Linker',
    'LinkerInstance',
    'Instance',
    'Func',
    'ResourceType',
    'ResourceAny',
    'ResourceHost',
    'ComponentType',
    'ModuleType',
    'ComponentItem',
    'ComponentInstanceType',
    'FuncType',
    'ValType',
    'Bool',
    'U8',
    'U16',
    'U32',
    'U64',
    'S8',
    'S16',
    'S32',
    'S64',
    'F32',
    'F64',
    'Char',
    'String',
    'ErrorContext',
    'ListType',
    'RecordType',
    'TupleType',
    'VariantType',
    'EnumType',
    'OptionType',
    'ResultType',
    'FlagsType',
    'StreamType',
    'FutureType',
    'OwnType',
    'BorrowType',
    'Variant',
    'Record',
]



# mypy: ignore-errors

# This is a small script to parse the header files from wasmtime and generate
# appropriate function definitions in Python for each exported function. This
# also reflects types into Python with `ctypes`. While there's at least one
# other generate that does this already it seemed to not quite fit our purposes
# with lots of extra an unnecessary boilerplate.

from pycparser import c_ast, parse_file
import sys

class Visitor(c_ast.NodeVisitor):
    def __init__(self):
        self.ret = ''
        self.ret += '# flake8: noqa\n'
        self.ret += '#\n'
        self.ret += '# This is a procedurally generated file, DO NOT EDIT\n'
        self.ret += '# instead edit `./ci/cbindgen.py` at the root of the repo\n'
        self.ret += '\n'
        self.ret += 'import ctypes\n'
        self.ret += 'from typing import Any\n'
        self.ret += 'from enum import Enum, auto\n'
        self.ret += 'from ._ffi import dll, wasm_val_t, wasm_ref_t\n'
        self.forward_declared = {}

    # Skip all function definitions, we don't bind those
    def visit_FuncDef(self, node):
        pass

    def visit_Struct(self, node):
        if not node.name or not node.name.startswith('was'):
            return

        # This is hand-generated since it has an anonymous union in it
        if node.name == 'wasm_val_t' or node.name == 'wasm_ref_t':
            return

        self.ret += "\n"
        if not node.decls:
            self.forward_declared[node.name] = True
            self.ret += "class {}(ctypes.Structure):\n".format(node.name)
            self.ret += "    pass\n"
            return

        anon_decl = 0
        for decl in node.decls:
            if not decl.name:
                assert(isinstance(decl.type, c_ast.Struct))
                decl.type.name = node.name + '_anon_' + str(anon_decl)
                self.visit_Struct(decl.type)
                anon_decl += 1
                decl.name = '_anon_' + str(anon_decl)

        if node.name in self.forward_declared:
            self.ret += "{}._fields_ = [\n".format(node.name)
        else:
            self.ret += "class {}(ctypes.Structure):\n".format(node.name)
            self.ret += "    _fields_ = [\n"

        for decl in node.decls:
            self.ret += "        (\"{}\", {}),\n".format(decl.name, type_name(decl.type))
        self.ret += "    ]\n"

        if not node.name in self.forward_declared:
            for decl in node.decls:
                self.ret += "    {}: {}\n".format(decl.name, type_name(decl.type, typing=True))

    def visit_Union(self, node):
        if not node.name or not node.name.startswith('was'):
            return
        assert(node.decls)

        self.ret += "\n"
        self.ret += "class {}(ctypes.Union):\n".format(node.name)
        self.ret += "    _fields_ = [\n"
        for decl in node.decls:
            self.ret += "        (\"{}\", {}),\n".format(name(decl.name), type_name(decl.type))
        self.ret += "    ]\n"
        for decl in node.decls:
            self.ret += "    {}: {}".format(name(decl.name), type_name(decl.type, typing=True))
            if decl.name == 'v128':
                self.ret += '  # type: ignore'
            self.ret += "\n"

    def visit_Enum(self, node):
        if not node.name or not node.name.startswith('was'):
            return

        self.ret += "\n"
        self.ret += "class {}(Enum):\n".format(node.name)
        for enumerator in node.values.enumerators:
            if enumerator.value:
                self.ret += "    {} = {}\n".format(enumerator.name, enumerator.value.value)
            else:
                self.ret += "    {} = auto()\n".format(enumerator.name)


    def visit_Typedef(self, node):
        if not node.name or not node.name.startswith('was'):
            return

        # Given anonymous structs in typedefs names by default.
        if isinstance(node.type, c_ast.TypeDecl):
            if isinstance(node.type.type, c_ast.Struct) or \
                isinstance(node.type.type, c_ast.Union):
                if node.type.type.name is None:
                    if node.name.endswith('_t'):
                        node.type.type.name = node.name[:-2]

        self.visit(node.type)
        tyname = type_name(node.type)
        if tyname != node.name:
            self.ret += "\n"
            if isinstance(node.type, c_ast.ArrayDecl):
                self.ret += "{} = {} * {}\n".format(node.name, type_name(node.type.type), node.type.dim.value)
            else:
                self.ret += "{} = {}\n".format(node.name, type_name(node.type))

    def visit_FuncDecl(self, node):
        if isinstance(node.type, c_ast.TypeDecl):
            ptr = False
            ty = node.type
        elif isinstance(node.type, c_ast.PtrDecl):
            ptr = True
            ty = node.type.type
        name = ty.declname
        # This is probably a type, skip it
        if name.endswith('_t'):
            return
        # Skip anything not related to wasi or wasm
        if not name.startswith('was'):
            return

        # This function forward-declares `wasm_instance_t` which doesn't work
        # with this binding generator, but for now this isn't used anyway so
        # just skip it.
        if name == 'wasm_frame_instance':
            return

        ret = ty.type

        argpairs = []
        argtypes = []
        argnames = []
        if node.args:
            for i, param in enumerate(node.args.params):
                argname = param.name
                if not argname or argname == "import" or argname == "global":
                    argname = "arg{}".format(i)
                tyname = type_name(param.type)
                if i == 0 and tyname == "None":
                    continue
                argpairs.append("{}: Any".format(argname))
                argnames.append(argname)
                argtypes.append(tyname)

        # It seems like this is the actual return value of the function, not a
        # pointer. Special-case this so the type-checking agrees with runtime.
        if type_name(ret, ptr) == 'c_void_p':
            retty = 'int'
        else:
            retty = type_name(node.type, ptr, typing=True)

        self.ret += "\n"
        self.ret += "_{0} = dll.{0}\n".format(name)
        self.ret += "_{}.restype = {}\n".format(name, type_name(ret, ptr))
        self.ret += "_{}.argtypes = [{}]\n".format(name, ', '.join(argtypes))
        self.ret += "def {}({}) -> {}:\n".format(name, ', '.join(argpairs), retty)
        self.ret += "    return _{}({})  # type: ignore\n".format(name, ', '.join(argnames))


def name(name):
    if name == 'global':
        return 'global_'
    return name


def type_name(ty, ptr=False, typing=False):
    while isinstance(ty, c_ast.TypeDecl):
        ty = ty.type

    if ptr:
        if typing:
            return "ctypes._Pointer"
        if isinstance(ty, c_ast.IdentifierType) and ty.names[0] == "void":
            return "ctypes.c_void_p"
        elif not isinstance(ty, c_ast.FuncDecl):
            return "ctypes.POINTER({})".format(type_name(ty, False, typing))

    if isinstance(ty, c_ast.IdentifierType):
        if ty.names == ['unsigned', 'char']:
            return "int" if typing else "ctypes.c_ubyte"
        assert(len(ty.names) == 1)

        if ty.names[0] == "void":
            return "None"
        elif ty.names[0] == "_Bool":
            return "bool" if typing else "ctypes.c_bool"
        elif ty.names[0] == "byte_t":
            return "ctypes.c_ubyte"
        elif ty.names[0] == "int8_t":
            return "ctypes.c_int8"
        elif ty.names[0] == "uint8_t":
            return "ctypes.c_uint8"
        elif ty.names[0] == "int16_t":
            return "ctypes.c_int16"
        elif ty.names[0] == "uint16_t":
            return "ctypes.c_uint16"
        elif ty.names[0] == "int32_t":
            return "int" if typing else "ctypes.c_int32"
        elif ty.names[0] == "uint32_t":
            return "int" if typing else "ctypes.c_uint32"
        elif ty.names[0] == "uint64_t":
            return "int" if typing else "ctypes.c_uint64"
        elif ty.names[0] == "int64_t":
            return "int" if typing else "ctypes.c_int64"
        elif ty.names[0] == "float32_t":
            return "float" if typing else "ctypes.c_float"
        elif ty.names[0] == "float64_t":
            return "float" if typing else "ctypes.c_double"
        elif ty.names[0] == "size_t":
            return "int" if typing else "ctypes.c_size_t"
        elif ty.names[0] == "ptrdiff_t":
            return "int" if typing else "ctypes.c_ssize_t"
        elif ty.names[0] == "char":
            return "ctypes.c_char"
        elif ty.names[0] == "int":
            return "int" if typing else "ctypes.c_int"
        # ctypes values can't stand as typedefs, so just use the pointer type here
        elif typing and 'callback_t' in ty.names[0]:
            return "ctypes._Pointer"
        elif typing and ('size' in ty.names[0] or 'pages' in ty.names[0]):
            return "int"
        return ty.names[0]
    elif isinstance(ty, c_ast.Struct):
        return ty.name
    elif isinstance(ty, c_ast.Union):
        return ty.name
    elif isinstance(ty, c_ast.Enum):
        return ty.name
    elif isinstance(ty, c_ast.FuncDecl):
        tys = []
        # TODO: apparently errors are thrown if we faithfully represent the
        # pointer type here, seems odd?
        if isinstance(ty.type, c_ast.PtrDecl):
            tys.append("ctypes.c_size_t")
        else:
            tys.append(type_name(ty.type))
        if ty.args and ty.args.params:
            for param in ty.args.params:
                tys.append(type_name(param.type))
        return "ctypes.CFUNCTYPE({})".format(', '.join(tys))
    elif isinstance(ty, c_ast.PtrDecl) or isinstance(ty, c_ast.ArrayDecl):
        return type_name(ty.type, True, typing)
    else:
        raise RuntimeError("unknown {}".format(ty))


def run():
    ast = parse_file(
        './wasmtime/include/wasmtime.h',
        use_cpp=True,
        cpp_path='gcc',
        cpp_args=[
            '-E',
            '-I./wasmtime/include',
            '-D__attribute__(x)=',
            '-D__asm__(x)=',
            '-D__asm(x)=',
            '-D__volatile__(x)=',
            '-D_Static_assert(x, y)=',
            '-Dstatic_assert(x, y)=',
            '-D__restrict=',
            '-D__restrict__=',
            '-D__extension__=',
            '-D__inline__=',
            '-D__signed=',
            '-D__builtin_va_list=int',
        ]
    )

    v = Visitor()
    v.visit(ast)
    return v.ret

if __name__ == "__main__":
    with open("wasmtime/_bindings.py", "w") as f:
        f.write(run())
elif sys.platform == 'linux':
    with open("wasmtime/_bindings.py", "r") as f:
        contents = f.read()
        if contents != run():
            raise RuntimeError("bindings need an update, run this script")



import unittest

from wasmtime import Store, WasmtimeError
from wasmtime.component import *


class TestValue(unittest.TestCase):
    def test_resource_type(self):
        r1 = ResourceType.host(1)
        r2 = ResourceType.host(2)
        r3 = ResourceType.host(1)
        self.assertNotEqual(r1, r2)
        self.assertEqual(r1, r3)

    def test_resource_host(self):
        store = Store()
        r1 = ResourceHost.own(42, 1)

        r2 = r1.to_any(store)
        self.assertEqual(r2.type, ResourceType.host(1))
        self.assertTrue(r2.owned)

        r3 = r2.to_host(store)
        self.assertTrue(r3.owned)
        self.assertEqual(r3.rep, 42)
        self.assertEqual(r3.type, 1)

        with self.assertRaises(WasmtimeError):
            r2.drop(store)

        r4 = r3.to_any(store)
        r4.drop(store)

        r5 = ResourceHost.borrow(84, 2)
        self.assertFalse(r5.owned)
        self.assertEqual(r5.rep, 84)
        self.assertEqual(r5.type, 2)

        with self.assertRaises(WasmtimeError):
            ResourceHost()

    def test_resource_host_dtor(self):
        store = Store()
        linker = Linker(store.engine)

        def drop(_store, rep):
            self.assertEqual(rep, 100)

        with linker.root() as l:
            l.add_resource('t', ResourceType.host(23), drop)

        component = Component(store.engine, """
            (component
                (import "t" (type $t (sub resource)))
                (core func $drop (canon resource.drop $t))
                (core module $a
                    (import "" "drop" (func $drop (param i32)))
                    (func (export "drop") (param i32)
                        local.get 0
                        call $drop)
                )
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "drop" (func $drop))
                    ))
                ))
                (func (export "drop") (param "x" (own $t))
                    (canon lift (core func $a "drop")))
            )
        """)
        instance = linker.instantiate(store, component)
        f = instance.get_func(store, 'drop')
        assert(f is not None)

        with self.assertRaises(WasmtimeError):
            f(store, ResourceHost.own(1, 2))

        instance = linker.instantiate(store, component)
        f = instance.get_func(store, 'drop')
        assert(f is not None)
        f(store, ResourceHost.own(100, 23))

    def test_exception_in_host_resource_dtor(self):
        store = Store()
        linker = Linker(store.engine)

        def drop(_store, _rep):
            raise RuntimeError('oh no')

        with linker.root() as l:
            l.add_resource('t', ResourceType.host(2), drop)

        component = Component(store.engine, """
            (component
                (import "t" (type $t (sub resource)))
                (core func $drop (canon resource.drop $t))
                (core module $a
                    (import "" "drop" (func $drop (param i32)))
                    (func (export "drop") (param i32)
                        local.get 0
                        call $drop)
                )
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "drop" (func $drop))
                    ))
                ))
                (func (export "drop") (param "x" (own $t))
                    (canon lift (core func $a "drop")))
            )
        """)
        instance = linker.instantiate(store, component)
        f = instance.get_func(store, 'drop')
        assert(f is not None)

        with self.assertRaises(RuntimeError) as cm:
            f(store, ResourceHost.own(1, 2))
        self.assertEqual(str(cm.exception), 'oh no')

    def test_resource_any(self):
        with self.assertRaises(WasmtimeError):
            ResourceAny()

        store = Store()
        linker = Linker(store.engine)

        component = Component(store.engine, """
            (component
                (type $t' (resource (rep i32)))
                (export $t "t" (type $t'))
                (core func $new (canon resource.new $t))
                (core func $drop (canon resource.drop $t))
                (core module $a
                    (import "" "new" (func $new (param i32) (result i32)))
                    (import "" "drop" (func $drop (param i32)))
                    (func (export "new") (param i32) (result i32)
                        local.get 0
                        call $new)
                    (func (export "drop") (param i32)
                        local.get 0
                        call $drop)
                )
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "new" (func $new))
                        (export "drop" (func $drop))
                    ))
                ))
                (func (export "new") (param "x" u32) (result (own $t))
                    (canon lift (core func $a "new")))
                (func (export "drop") (param "x" (own $t))
                    (canon lift (core func $a "drop")))
            )
        """)
        instance = linker.instantiate(store, component)
        new = instance.get_func(store, 'new')
        drop = instance.get_func(store, 'drop')
        assert(new is not None)
        assert(drop is not None)

        r1 = new(store, 100)
        new.post_return(store)
        r2 = new(store, 200)
        new.post_return(store)

        with self.assertRaises(WasmtimeError):
            r1.to_host(store)

        r1.drop(store)
        drop(store, r2)
        drop.post_return(store)

        with self.assertRaises(WasmtimeError):
            drop(store, ResourceHost.own(1, 1))



from wasmtime import Store, Module, Instance, Func, FuncType

def run():
    # Almost all operations in wasmtime require a contextual "store" argument to be
    # shared amongst objects
    store = Store()

    # Here we can compile a `Module` which is then ready for instantiation
    # afterwards
    module = Module.from_file(store.engine, './examples/hello.wat')

    # Our module needs one import, so we'll create that here.


    def say_hello():
        print("Hello from Python!")


    hello = Func(store, FuncType([], []), say_hello)

    # And with all that we can instantiate our module and call the export!
    instance = Instance(store, module, [hello])
    instance.exports(store)["run"](store)

if __name__ == '__main__':
    run()



import typing

if typing.TYPE_CHECKING:
    from ._func import Func
    from ._globals import Global
    from ._memory import Memory
    from ._sharedmemory import SharedMemory
    from ._table import Table

AsExtern = typing.Union["Func", "Table", "Memory", "SharedMemory",  "Global"]



import unittest

from wasmtime import *


class TestTrap(unittest.TestCase):
    def test_new(self):
        trap = Trap('x')
        self.assertEqual(trap.message, u'x')

    def test_frames(self):
        store = Store()
        module = Module(store.engine, """
            (module $module
                (func (export "init")
                    call $foo)
                (func $foo
                    call $bar)
                (func $bar
                    unreachable)
            )
        """)
        i = Instance(store, module, [])
        with self.assertRaises(Trap) as exn:
            e = i.exports(store).by_index[0]
            assert(isinstance(e, Func))
            e(store)
        trap = exn.exception

        frames = trap.frames
        self.assertEqual(len(frames), 3)
        self.assertEqual(frames[0].func_index, 2)
        self.assertEqual(frames[1].func_index, 1)
        self.assertEqual(frames[2].func_index, 0)

        self.assertEqual(frames[0].func_name, 'bar')
        self.assertEqual(frames[1].func_name, 'foo')
        self.assertEqual(frames[2].func_name, None)

        self.assertEqual(frames[0].module_name, 'module')
        self.assertEqual(frames[1].module_name, 'module')
        self.assertEqual(frames[2].module_name, 'module')

        self.assertEqual(str(trap), """\
error while executing at wasm backtrace:
    0:     0x2d - module!bar
    1:     0x28 - module!foo
    2:     0x23 - module!<wasm function 0>

Caused by:
    wasm trap: wasm `unreachable` instruction executed\
""")
        self.assertEqual(trap.trap_code, TrapCode.UNREACHABLE)

    def test_frames_no_module(self):
        store = Store()
        module = Module(store.engine, """
            (module
                (func (export "init") unreachable)
            )
        """)
        i = Instance(store, module, [])
        with self.assertRaises(Trap) as exn:
            e = i.exports(store).by_index[0]
            assert(isinstance(e, Func))
            e(store)
        trap = exn.exception

        frames = trap.frames
        self.assertEqual(len(frames), 1)
        self.assertEqual(frames[0].func_index, 0)
        self.assertEqual(frames[0].func_name, None)
        self.assertEqual(frames[0].module_name, None)

    def test_wasi_exit(self):
        linker = Linker(Engine())
        linker.define_wasi()
        module = Module(linker.engine, """
            (module
                (import "wasi_snapshot_preview1" "proc_exit" (func $exit (param i32)))
                (memory (export "memory") 1)
                (func (export "exit") (param i32)
                    local.get 0
                    call $exit)
            )
        """)
        store = Store(linker.engine)
        store.set_wasi(WasiConfig())
        instance = linker.instantiate(store, module)
        exit = instance.exports(store)["exit"]
        assert(isinstance(exit, Func))

        with self.assertRaises(ExitTrap) as exn:
            exit(store, 0)
        self.assertEqual(exn.exception.code, 0)
        self.assertRegex(str(exn.exception), 'Exited with i32 exit status 0')

        with self.assertRaises(ExitTrap) as exn:
            exit(store, 1)
        self.assertEqual(exn.exception.code, 1)



import unittest

from wasmtime import *


class TestValue(unittest.TestCase):
    def test_i32(self):
        with self.assertRaises(TypeError):
            Val.i32('')  # type: ignore
        with self.assertRaises(TypeError):
            Val.i32(1.2)  # type: ignore

        i = Val.i32(1)
        self.assertEqual(i.as_i32(), 1)
        i = Val.i32(0xffffffff)
        self.assertEqual(i.as_i32(), -1)
        i = Val.i32(0x800000000)
        self.assertEqual(i.as_i32(), 0)

        self.assertEqual(None, i.as_i64())
        self.assertEqual(None, i.as_f32())
        self.assertEqual(None, i.as_f64())
        self.assertEqual(i.type, ValType.i32())

    def test_i64(self):
        with self.assertRaises(TypeError):
            Val.i64('')  # type: ignore
        with self.assertRaises(TypeError):
            Val.i64(1.2)  # type: ignore

        i = Val.i64(1)
        self.assertEqual(i.as_i64(), 1)
        i = Val.i64(0xffffffff)
        self.assertEqual(i.as_i64(), 0xffffffff)
        i = Val.i64(0x800000000)
        self.assertEqual(i.as_i64(), 0x800000000)
        i = Val.i64(0xffffffffffffffff)
        self.assertEqual(i.as_i64(), -1)
        i = Val.i64(0x80000000000000000)
        self.assertEqual(i.as_i64(), 0)

        self.assertEqual(None, i.as_i32())
        self.assertEqual(None, i.as_f32())
        self.assertEqual(None, i.as_f64())
        self.assertEqual(i.type, ValType.i64())

    def test_f32(self):
        with self.assertRaises(TypeError):
            Val.f32('')  # type: ignore
        with self.assertRaises(TypeError):
            Val.f32(1)

        i = Val.f32(1.0)
        self.assertEqual(i.as_f32(), 1.0)
        self.assertEqual(None, i.as_i64())
        self.assertEqual(None, i.as_i32())
        self.assertEqual(None, i.as_f64())
        self.assertEqual(i.type, ValType.f32())

    def test_f64(self):
        with self.assertRaises(TypeError):
            Val.f64('')  # type: ignore
        with self.assertRaises(TypeError):
            Val.f64(1)

        i = Val.f64(1.0)
        self.assertEqual(i.as_f64(), 1.0)
        self.assertEqual(None, i.as_i32())
        self.assertEqual(None, i.as_i64())
        self.assertEqual(None, i.as_f32())
        self.assertEqual(i.type, ValType.f64())



import unittest
import tempfile

from wasmtime import *


class TestModule(unittest.TestCase):
    def test_smoke(self):
        Module(Engine(), '(module)')
        Module(Engine(), bytes(b'\0asm\x01\0\0\0'))
        Module(Engine(), bytearray(b'\0asm\x01\0\0\0'))

    def test_invalid(self):
        with self.assertRaises(AttributeError):
            Module.validate(1, b'')  # type: ignore
        with self.assertRaises(TypeError):
            Module.validate(Store(), 2)  # type: ignore
        with self.assertRaises(TypeError):
            Module(1, b'')  # type: ignore
        with self.assertRaises(TypeError):
            Module(Engine(), 2)  # type: ignore
        with self.assertRaises(WasmtimeError):
            Module(Engine(), b'')
        with self.assertRaises(WasmtimeError):
            Module(Engine(), b'\x00')

    def test_validate(self):
        engine = Engine()
        Module.validate(engine, b'\0asm\x01\0\0\0')
        with self.assertRaises(WasmtimeError):
            Module.validate(engine, b'')

    def test_imports(self):
        store = Store()
        module = Module(store.engine, '(module)')
        self.assertEqual(module.imports, [])

        module = Module(store.engine, """
            (module
                (import "" "" (func))
                (import "a" "bcd" (global i32))
                (import "" "" (memory 1))
                (import "" "x" (table 1 funcref))
            )
        """)
        imports = module.imports
        self.assertEqual(len(imports), 4)
        self.assertEqual(imports[0].module, "")
        self.assertEqual(imports[0].name, "")
        ty = imports[0].type
        assert(isinstance(ty, FuncType))
        self.assertEqual(ty.params, [])
        self.assertEqual(ty.results, [])

        self.assertEqual(imports[1].module, "a")
        self.assertEqual(imports[1].name, "bcd")
        ty = imports[1].type
        assert(isinstance(ty, GlobalType))
        self.assertEqual(ty.content, ValType.i32())
        self.assertFalse(ty.mutable)

        self.assertEqual(imports[2].module, "")
        self.assertEqual(imports[2].name, "")
        ty = imports[2].type
        assert(isinstance(ty, MemoryType))
        self.assertEqual(ty.limits, Limits(1, None))

        self.assertEqual(imports[3].module, "")
        self.assertEqual(imports[3].name, "x")
        ty = imports[3].type
        assert(isinstance(ty, TableType))
        self.assertEqual(ty.limits, Limits(1, None))
        self.assertEqual(ty.element, ValType.funcref())

    def test_exports(self):
        store = Store()
        module = Module(store.engine, '(module)')
        self.assertEqual(module.exports, [])

        module = Module(store.engine, """
            (module
                (func (export "a") (param i32 f32) (result f64)
                    f64.const 0)
                (global (export "") (mut i32) (i32.const 1))
                (memory (export "mem") 1)
                (table (export "table") 1 funcref)
            )
        """)
        exports = module.exports
        self.assertEqual(len(exports), 4)
        self.assertEqual(exports[0].name, "a")
        ty = exports[0].type
        assert(isinstance(ty, FuncType))
        self.assertEqual(ty.params, [ValType.i32(), ValType.f32()])
        self.assertEqual(ty.results, [ValType.f64()])

        self.assertEqual(exports[1].name, "")
        ty = exports[1].type
        assert(isinstance(ty, GlobalType))
        self.assertEqual(ty.content, ValType.i32())
        self.assertTrue(ty.mutable)

        self.assertEqual(exports[2].name, "mem")
        ty = exports[2].type
        assert(isinstance(ty, MemoryType))
        self.assertEqual(ty.limits, Limits(1, None))

        self.assertEqual(exports[3].name, "table")
        ty = exports[3].type
        assert(isinstance(ty, TableType))
        self.assertEqual(ty.limits, Limits(1, None))
        self.assertEqual(ty.element, ValType.funcref())

    def test_serialize(self):
        engine = Engine()
        module = Module(engine, '(module)')
        encoded = module.serialize()
        module = Module.deserialize(engine, encoded)
        assert(len(module.imports) == 0)
        assert(len(module.exports) == 0)
        with tempfile.TemporaryDirectory() as d:
            path = d + '/module.bin'
            with open(path, 'wb') as f:
                f.write(encoded)

            # Run the destructor for `Module` which has an mmap to the file
            # which prevents deletion on Windows.
            with Module.deserialize_file(engine, path) as module:
                assert(len(module.imports) == 0)
                assert(len(module.exports) == 0)



from ._error import WasmtimeError
from ._types import ValType
import ctypes
import typing
from ._slab import Slab

from . import _ffi as ffi

if typing.TYPE_CHECKING:
    from ._store import Storelike


SLAB: Slab[typing.Any] = Slab()


# Note that 0 as a `c_void_p` is reserved for null-lookalike externrefs so
# indices from the slabs are off-by-one.
@ctypes.CFUNCTYPE(None, ctypes.c_void_p)
def _externref_finalizer(idx: int) -> None:
    SLAB.deallocate(idx - 1)


def _intern(obj: typing.Any) -> int:
    return SLAB.allocate(obj) + 1


def _unintern(idx: int) -> typing.Any:
    return SLAB.get(idx - 1)


class Val:
    _kind: ffi.wasmtime_valkind_t
    _val: typing.Any

    @classmethod
    def i32(cls, val: int) -> "Val":
        """
        Create a new 32-bit integer value
        """
        if not isinstance(val, int):
            raise TypeError("expected an integer")
        val = ffi.wasmtime_valunion_t(i32=val).i32
        return Val(ffi.WASMTIME_I32, val)

    @classmethod
    def i64(cls, val: int) -> "Val":
        """
        Create a new 64-bit integer value
        """
        if not isinstance(val, int):
            raise TypeError("expected an integer")
        val = ffi.wasmtime_valunion_t(i64=val).i64
        return Val(ffi.WASMTIME_I64, val)

    @classmethod
    def f32(cls, val: float) -> "Val":
        """
        Create a new 32-bit float value
        """
        if not isinstance(val, float):
            raise TypeError("expected a float")
        val = ffi.wasmtime_valunion_t(f32=val).f32
        return Val(ffi.WASMTIME_F32, val)

    @classmethod
    def f64(cls, val: float) -> "Val":
        """
        Create a new 64-bit float value
        """
        if not isinstance(val, float):
            raise TypeError("expected a float")
        val = ffi.wasmtime_valunion_t(f64=val).f64
        return Val(ffi.WASMTIME_F64, val)

    @classmethod
    def externref(cls, extern: typing.Optional[typing.Any]) -> "Val":
        return Val(ffi.WASMTIME_EXTERNREF, extern)

    @classmethod
    def funcref(cls, f: "typing.Optional[wasmtime.Func]") -> "Val":
        return Val(ffi.WASMTIME_FUNCREF, f)

    @classmethod
    def ref_null(cls, ty: ValType) -> "Val":
        """
        Create a null reference value of the given type.

        Raise an exception if `ty` is not a reference type.
        """
        if ty == ValType.externref():
            return Val.externref(None)
        if ty == ValType.funcref():
            return Val.funcref(None)
        raise WasmtimeError("Invalid reference type for `ref_null`: %s" % ty)

    def __init__(self, kind: ffi.wasmtime_valkind_t, val: typing.Any):
        self._kind = kind
        self._val = val

    def __eq__(self, rhs: typing.Any) -> typing.Any:
        if isinstance(rhs, Val):
            return self._kind == rhs._kind and self._val == rhs._val
        return self._val == rhs

    @classmethod
    def _convert_to_raw(cls, store: 'Storelike', ty: ValType, val: typing.Any) -> ffi.wasmtime_val_t:
        if isinstance(val, Val):
            if ty != val.type:
                raise TypeError("wrong type of `Val` provided")
            return val._new_raw(store)
        if ty == ValType.externref():
            return Val.externref(val)._new_raw(store)
        elif isinstance(val, int):
            if ty == ValType.i32():
                return Val.i32(val)._new_raw(store)
            if ty == ValType.i64():
                return Val.i64(val)._new_raw(store)
        elif isinstance(val, float):
            if ty == ValType.f32():
                return Val.f32(val)._new_raw(store)
            if ty == ValType.f64():
                return Val.f64(val)._new_raw(store)
        elif isinstance(val, wasmtime.Func):
            return Val.funcref(val)._new_raw(store)
        elif val is None:
            if ty == ValType.externref():
                return Val.externref(None)._new_raw(store)
            if ty == ValType.funcref():
                return Val.funcref(None)._new_raw(store)
        raise TypeError("don't know how to convert %r to %s" % (val, ty))

    def _new_raw(self, store: 'Storelike') -> ffi.wasmtime_val_t:
        ret = ffi.wasmtime_val_t(kind = self._kind)
        if ret.kind == ffi.WASMTIME_I32.value:
            ret.of.i32 = self._val
        elif ret.kind == ffi.WASMTIME_I64.value:
            ret.of.i64 = self._val
        elif ret.kind == ffi.WASMTIME_F32.value:
            ret.of.f32 = self._val
        elif ret.kind == ffi.WASMTIME_F64.value:
            ret.of.f64 = self._val
        elif ret.kind == ffi.WASMTIME_EXTERNREF.value:
            if self._val is not None:
                extern_id = _intern(self._val)
                if not ffi.wasmtime_externref_new(store._context(), extern_id, _externref_finalizer,
                                              ctypes.byref(ret.of.externref)):
                    raise WasmtimeError("failed to create an externref value")
            else:

                ret.of.externref.store_id = 0
        elif ret.kind == ffi.WASMTIME_FUNCREF.value:
            if self._val is not None:
                ret.of.funcref = self._val._func
        else:
            raise RuntimeError("unknown kind")
        return ret

    @classmethod
    def _from_raw(cls, store: 'Storelike', raw: ffi.wasmtime_val_t, owned: bool = True) -> 'Val':
        val: typing.Any = None
        if raw.kind == ffi.WASMTIME_I32.value:
            val = raw.of.i32
        elif raw.kind == ffi.WASMTIME_I64.value:
            val = raw.of.i64
        elif raw.kind == ffi.WASMTIME_F32.value:
            val = raw.of.f32
        elif raw.kind == ffi.WASMTIME_F64.value:
            val = raw.of.f64
        elif raw.kind == ffi.WASMTIME_EXTERNREF.value:
            if raw.of.externref:
                extern_id = ffi.wasmtime_externref_data(store._context(), raw.of.externref)
                # FIXME https://github.com/bytecodealliance/wasmtime-py/issues/303
                if extern_id:
                    val = _unintern(extern_id)  # type: ignore
        elif raw.kind == ffi.WASMTIME_FUNCREF.value:
            if raw.of.funcref.store_id != 0:
                val = wasmtime.Func._from_raw(raw.of.funcref)
        else:
            raise WasmtimeError("Unkown `ffi.wasmtime_valkind_t`: {}".format(raw.kind))

        if owned:
            ffi.wasmtime_val_unroot(ctypes.byref(raw))

        return Val(raw.kind, val)

    @property
    def value(self) -> typing.Any:
        """
        Get the the underlying value as a python value
        """
        return self._val

    def as_i32(self) -> typing.Optional[int]:
        """
        Get the 32-bit integer value of this value, or `None` if it's not an i32
        """
        if self._kind == ffi.WASMTIME_I32:
            assert(isinstance(self._val, int))
            return self._val
        else:
            return None

    def as_i64(self) -> typing.Optional[int]:
        """
        Get the 64-bit integer value of this value, or `None` if it's not an i64
        """
        if self._kind == ffi.WASMTIME_I64:
            assert(isinstance(self._val, int))
            return self._val
        else:
            return None

    def as_f32(self) -> typing.Optional[float]:
        """
        Get the 32-bit float value of this value, or `None` if it's not an f32
        """
        if self._kind == ffi.WASMTIME_F32:
            assert(isinstance(self._val, float))
            return self._val
        else:
            return None

    def as_f64(self) -> typing.Optional[float]:
        """
        Get the 64-bit float value of this value, or `None` if it's not an f64
        """
        if self._kind == ffi.WASMTIME_F64:
            assert(isinstance(self._val, float))
            return self._val
        else:
            return None

    def as_externref(self) -> typing.Optional[typing.Any]:
        """
        Get the extern data referenced by this `externref` value, or `None` if
        it's not an `externref`.
        """
        if self._kind == ffi.WASMTIME_EXTERNREF:
            return self._val
        else:
            return None

    def as_funcref(self) -> typing.Optional["wasmtime.Func"]:
        """
        Get the function that this `funcref` value is referencing, or `None` if
        this is not a `funcref` value, or is a null reference.
        """
        if self._kind == ffi.WASMTIME_FUNCREF:
            assert(isinstance(self._val, wasmtime.Func))
            return self._val
        else:
            return None

    @property
    def type(self) -> ValType:
        """
        Returns the `ValType` corresponding to this `Val`
        """
        kind = self._kind
        if kind == ffi.WASMTIME_I32:
            return ValType.i32()
        elif kind == ffi.WASMTIME_I64:
            return ValType.i64()
        elif kind == ffi.WASMTIME_F32:
            return ValType.f32()
        elif kind == ffi.WASMTIME_F64:
            return ValType.f64()
        elif kind == ffi.WASMTIME_V128:
            raise Exception("unimplemented v128 type")
        elif kind == ffi.WASMTIME_EXTERNREF:
            return ValType.externref()
        elif kind == ffi.WASMTIME_FUNCREF:
            return ValType.funcref()
        else:
            raise Exception("unknown kind %d" % kind.value)


# This needs to be imported so that mypy understands `wasmtime.Func` in typings,
# but it also has to come after `Val`'s definition to deal with circular module
# dependency issues.
import wasmtime  # noqa: E402



# An example of how to interact with wasm memory.
#
# Here a small wasm module is used to show how memory is initialized, how to
# read and write memory through the `Memory` object, and how wasm functions
# can trap when dealing with out-of-bounds addresses.

from wasmtime import Store, Module, Instance, Trap, MemoryType, Memory, Limits, WasmtimeError

def run():
    # Create our `Store` context and then compile a module and create an
    # instance from the compiled module all in one go.
    store = Store()
    module = Module.from_file(store.engine, "examples/memory.wat")
    instance = Instance(store, module, [])

    # Load up our exports from the instance
    exports = instance.exports(store)
    memory = exports["memory"]
    size_fn = exports["size"]
    load_fn = exports["load"]
    store_fn = exports["store"]

    print("Checking memory...")
    assert(memory.size(store) == 2)
    assert(memory.data_len(store) == 0x20000)

    # Note that usage of `data_ptr` is unsafe! This is a raw C pointer which is not
    # bounds checked at all. We checked our `data_len` above but you'll want to be
    # very careful when accessing data through `data_ptr()`
    assert(memory.data_ptr(store)[0] == 0)
    assert(memory.data_ptr(store)[0x1000] == 1)
    assert(memory.data_ptr(store)[0x1003] == 4)

    assert(size_fn(store) == 2)
    assert(load_fn(store, 0) == 0)
    assert(load_fn(store, 0x1000) == 1)
    assert(load_fn(store, 0x1003) == 4)
    assert(load_fn(store, 0x1ffff) == 0)


    def assert_traps(func):
        try:
            func()
            assert(False)
        except Trap:
            pass
        except WasmtimeError:
            pass


    # out of bounds trap
    assert_traps(lambda: load_fn(store, 0x20000))

    print("Mutating memory...")
    memory.data_ptr(store)[0x1003] = 5
    store_fn(store, 0x1002, 6)
    # out of bounds trap
    assert_traps(lambda: store_fn(store, 0x20000, 0))

    assert(memory.data_ptr(store)[0x1002] == 6)
    assert(memory.data_ptr(store)[0x1003] == 5)
    assert(load_fn(store, 0x1002) == 6)
    assert(load_fn(store, 0x1003) == 5)

    # Grow memory.
    print("Growing memory...")
    assert(memory.grow(store, 1))
    assert(memory.size(store) == 3)
    assert(memory.data_len(store) == 0x30000)

    assert(load_fn(store, 0x20000) == 0)
    store_fn(store, 0x20000, 0)
    assert_traps(lambda: load_fn(store, 0x30000))
    assert_traps(lambda: store_fn(store, 0x30000, 0))

    # Memory can fail to grow
    assert_traps(lambda: memory.grow(store, 1))
    assert(memory.grow(store, 0))

    print("Creating stand-alone memory...")
    memorytype = MemoryType(Limits(5, 5))
    memory2 = Memory(store, memorytype)
    assert(memory2.size(store) == 5)
    assert_traps(lambda: memory2.grow(store, 1))
    assert(memory2.grow(store, 0))

if __name__ == '__main__':
    run()



from . import _ffi as ffi
import ctypes
from typing import Optional, Any
from wasmtime import MemoryType, WasmtimeError, Engine, Managed
from ._store import Storelike



class SharedMemory(Managed["ctypes._Pointer[ffi.wasmtime_sharedmemory_t]"]):
    def __init__(self, engine: Engine, ty: MemoryType):
        """
        Creates a new shared memory in `store` with the given `ty`
        """

        sharedmemory_ptr = ctypes.POINTER(ffi.wasmtime_sharedmemory_t)()
        error = ffi.wasmtime_sharedmemory_new(engine.ptr(), ty.ptr(), ctypes.byref(sharedmemory_ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        self._set_ptr(sharedmemory_ptr)

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_sharedmemory_t]") -> None:
        ffi.wasmtime_sharedmemory_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_sharedmemory_t]") -> "SharedMemory":
        if not isinstance(ptr, ctypes.POINTER(ffi.wasmtime_sharedmemory_t)):
            raise TypeError("wrong shared memory pointer type provided to _from_ptr")
        
        ty: "SharedMemory" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    def type(self) -> MemoryType:
        """
        Gets the type of this memory as a `MemoryType`
        """

        ptr = ffi.wasmtime_sharedmemory_type(self.ptr())
        return MemoryType._from_ptr(ptr, None)

    def grow(self, delta: int) -> int:
        """
        Grows this memory by the given number of pages
        """

        if delta < 0:
            raise WasmtimeError("cannot grow by negative amount")
        prev = ctypes.c_uint64(0)
        error = ffi.wasmtime_sharedmemory_grow(self.ptr(), delta, ctypes.byref(prev))
        if error:
            raise WasmtimeError._from_ptr(error)
        return prev.value

    def size(self) -> int:
        """
        Returns the size, in WebAssembly pages, of this shared memory.
        """

        return ffi.wasmtime_sharedmemory_size(ctypes.byref(self.ptr()))

    def data_ptr(self) -> "ctypes._Pointer[ctypes.c_ubyte]":
        """
        Returns the raw pointer in memory where this wasm shared memory lives.

        Remember that all accesses to wasm shared memory should be bounds-checked
        against the `data_len` method.
        """
        return ffi.wasmtime_sharedmemory_data(self.ptr())

    def data_len(self) -> int:
        """
        Returns the raw byte length of this memory.
        """

        return ffi.wasmtime_sharedmemory_data_size(self.ptr())

    def _as_extern(self) -> ffi.wasmtime_extern_t:
        union = ffi.wasmtime_extern_union(sharedmemory=ctypes.pointer(self.ptr()))
        return ffi.wasmtime_extern_t(ffi.WASMTIME_EXTERN_SHAREDMEMORY, union)



import unittest

from wasmtime import *


class TestWat2Wasm(unittest.TestCase):
    def test_wat2wasm(self):
        wasm = wat2wasm('(module)')
        self.assertEqual(wasm, b'\0asm\x01\0\0\0')

    def test_errors(self):
        with self.assertRaises(WasmtimeError):
            wat2wasm('x')



import unittest
from contextlib import closing

from wasmtime import *


class TestConfig(unittest.TestCase):
    def test_smoke(self):
        config = Config()
        config.debug_info = True
        config.wasm_threads = True
        config.wasm_tail_call = True
        config.wasm_reference_types = True
        config.wasm_simd = True
        config.wasm_bulk_memory = True
        config.wasm_multi_value = True
        config.wasm_multi_memory = True
        config.wasm_memory64 = True
        config.wasm_exceptions = True
        config.cranelift_debug_verifier = True
        config.strategy = "cranelift"
        config.strategy = "auto"
        config.cache = True
        config.parallel_compilation = False
        with self.assertRaises(WasmtimeError):
            config.cache = "./test.toml"
        with self.assertRaises(WasmtimeError):
            config.strategy = "nonexistent-strategy"
        config.cranelift_opt_level = "none"
        config.cranelift_opt_level = "speed_and_size"
        config.cranelift_opt_level = "speed"
        with self.assertRaises(WasmtimeError):
            config.cranelift_opt_level = "nonexistent-level"
        config.profiler = "none"
        with self.assertRaises(WasmtimeError):
            config.profiler = "nonexistent-profiler"
        config.consume_fuel = True
        config.wasm_relaxed_simd = True
        config.wasm_relaxed_simd_deterministic = True

        with closing(config) as config:
            pass

        config.close()

        with self.assertRaises(ValueError):
            Engine(config)

        with self.assertRaises(ValueError):
            config.cache = True



import unittest

from wasmtime import *


class TestMemory(unittest.TestCase):
    def test_new(self):
        store = Store()
        ty = MemoryType(Limits(1, None))
        assert(not ty.is_64)
        memory = Memory(store, ty)
        self.assertEqual(memory.type(store).limits, Limits(1, None))
        self.assertEqual(memory.size(store), 1)
        self.assertTrue(memory.grow(store, 1))
        self.assertEqual(memory.size(store), 2)
        self.assertTrue(memory.grow(store, 0))
        self.assertEqual(memory.size(store), 2)
        with self.assertRaises(TypeError):
            memory.grow(store, '')  # type: ignore
        with self.assertRaises(WasmtimeError):
            memory.grow(store, -1)
        self.assertEqual(memory.data_ptr(store)[0], 0)
        self.assertEqual(memory.data_len(store), 65536 * 2)
        self.assertTrue(isinstance(memory.type(store), MemoryType))

    def test_grow(self):
        store = Store()
        ty = MemoryType(Limits(1, 2))
        memory = Memory(store, ty)
        assert(memory.grow(store, 1) == 1)
        assert(memory.grow(store, 0) == 2)
        with self.assertRaises(WasmtimeError):
            memory.grow(store, 1)

    def test_errors(self):
        store = Store()
        ty = MemoryType(Limits(1, 2))
        with self.assertRaises(AttributeError):
            Memory(1, ty)  # type: ignore
        with self.assertRaises(AttributeError):
            Memory(store, 1)  # type: ignore

    def test_large(self):
        ty = MemoryType(Limits(0x100000000, None), is_64=True)
        assert(ty.limits.min == 0x100000000)
        assert(ty.limits.max is None)
        assert(ty.is_64)
        ty = MemoryType(Limits(0x100000000, 0x100000000), is_64=True)
        assert(ty.limits.min == 0x100000000)
        assert(ty.limits.max == 0x100000000)

        with self.assertRaises(WasmtimeError):
            MemoryType(Limits(0x100000000, None))
        with self.assertRaises(WasmtimeError):
            MemoryType(Limits(1, 0x100000000))

    def test_slices(self):
        store = Store()
        ty = MemoryType(Limits(1, None))
        memory = Memory(store, ty)
        memory.grow(store, 2)
        data_ptr = memory.data_ptr(store)
        ba = bytearray([i for i in range(200)])
        size_bytes = memory.data_len(store)
        # happy cases
        offset = 2048
        ba_size = len(ba)
        # write with start and ommit stop
        memory.write(store, ba, offset)
        # check write success byte by byte, whole is asserted with read
        self.assertEqual(data_ptr[offset], 0)
        self.assertEqual(data_ptr[offset + 1], 1)
        self.assertEqual(data_ptr[offset + 199], 199)
        # read while and assert whole area
        out = memory.read(store, offset, offset + ba_size)
        self.assertEqual(ba, out)
        self.assertEqual(len(memory.read(store, -10)), 10)
        # write with start and stop
        memory.write(store, ba, offset + ba_size)
        out = memory.read(store, offset + ba_size, offset + ba_size + ba_size)
        self.assertEqual(ba, out)
        # assert old
        self.assertEqual(data_ptr[offset], 0)
        self.assertEqual(data_ptr[offset + 1], 1)
        self.assertEqual(data_ptr[offset + 199], 199)
        # assert new
        self.assertEqual(data_ptr[offset + ba_size], 0)
        self.assertEqual(data_ptr[offset + ba_size + 1], 1)
        self.assertEqual(data_ptr[offset + ba_size + 199], 199)
        # edge cases
        # empty slices
        self.assertEqual(len(memory.read(store, 0, 0)), 0)
        self.assertEqual(len(memory.read(store, offset, offset)), 0)
        self.assertEqual(len(memory.read(store, offset, offset - 1)), 0)
        # out of bound access returns empty array similar to list slice
        self.assertEqual(len(memory.read(store, size_bytes + 1)), 0)
        # write empty
        self.assertEqual(memory.write(store, bytearray(0), offset), 0)
        self.assertEqual(memory.write(store, bytearray(b""), offset), 0)
        with self.assertRaises(IndexError):
            memory.write(store, ba, size_bytes)
        with self.assertRaises(IndexError):
            memory.write(store, ba, size_bytes - ba_size + 1)
        self.assertEqual(memory.write(store, ba, -ba_size), ba_size)
        out = memory.read(store, -ba_size)
        self.assertEqual(ba, out)



import unittest

from wasmtime import *


class TestInstance(unittest.TestCase):
    def test_smoke(self):
        store = Store()
        module = Module(store.engine, '(module)')
        Instance(store, module, [])

    def test_export_func(self):
        store = Store()
        module = Module(store.engine, '(module (func (export "")))')
        instance = Instance(store, module, [])
        self.assertEqual(len(instance.exports(store)), 1)
        extern = instance.exports(store).by_index[0]
        assert(isinstance(extern, Func))
        assert(isinstance(extern.type(store), FuncType))

        extern(store)

        assert(instance.exports(store)[''] is not None)
        with self.assertRaises(KeyError):
            instance.exports(store)['x']
        with self.assertRaises(IndexError):
            instance.exports(store).by_index[100]
        assert(instance.exports(store).get('x') is None)

    def test_export_global(self):
        store = Store()
        module = Module(
            store.engine, '(module (global (export "") i32 (i32.const 3)))')
        instance = Instance(store, module, [])
        self.assertEqual(len(instance.exports(store)), 1)
        extern = instance.exports(store).by_index[0]
        assert(isinstance(extern, Global))
        self.assertEqual(extern.value(store), 3)
        assert(isinstance(extern.type(store), GlobalType))

    def test_export_memory(self):
        store = Store()
        module = Module(store.engine, '(module (memory (export "") 1))')
        instance = Instance(store, module, [])
        self.assertEqual(len(instance.exports(store)), 1)
        extern = instance.exports(store).by_index[0]
        assert(isinstance(extern, Memory))
        self.assertEqual(extern.size(store), 1)

    def test_export_table(self):
        store = Store()
        module = Module(store.engine, '(module (table (export "") 1 funcref))')
        instance = Instance(store, module, [])
        self.assertEqual(len(instance.exports(store)), 1)
        extern = instance.exports(store).by_index[0]
        assert(isinstance(extern, Table))

    def test_multiple_exports(self):
        store = Store()
        module = Module(store.engine, """
            (module
                (func (export "a"))
                (func (export "b"))
                (global (export "c") i32 (i32.const 0))
            )
        """)
        instance = Instance(store, module, [])
        exports = instance.exports(store)
        self.assertEqual(len(exports), 3)
        assert isinstance(exports.by_index[0], Func)
        assert isinstance(exports.by_index[1], Func)
        assert isinstance(exports.by_index[2], Global)
        # Test that exports acts like a normal map
        assert "a" in exports
        assert "b" in exports
        assert "d" not in exports
        assert set(exports) == {"a", "b", "c"}
        assert set(exports.values()) == set(exports.by_index)
        assert exports.get("d", 7) == 7
        assert isinstance(exports.get("b", 7), Func)

    def test_import_func(self):
        store = Store()
        module = Module(store.engine, """
            (module
                (import "" "" (func))
                (start 0)
            )
        """)
        hit = []
        func = Func(store, FuncType([], []), lambda: hit.append(True))
        Instance(store, module, [func])
        assert(len(hit) == 1)
        Instance(store, module, [func])
        assert(len(hit) == 2)

    def test_import_global(self):
        store = Store()
        module = Module(store.engine, """
            (module
                (import "" "" (global (mut i32)))
                (func (export "") (result i32)
                    global.get 0)
                (func (export "update")
                    i32.const 5
                    global.set 0)
            )
        """)
        g = Global(store, GlobalType(ValType.i32(), True), 2)
        instance = Instance(store, module, [g])
        f = instance.exports(store).by_index[0]
        assert(isinstance(f, Func))

        self.assertEqual(f(store), 2)
        g.set_value(store, 4)
        self.assertEqual(f(store), 4)

        instance2 = Instance(store, module, [g])
        f2 = instance2.exports(store).by_index[0]
        assert(isinstance(f2, Func))
        self.assertEqual(f(store), 4)
        self.assertEqual(f2(store), 4)

        update = instance.exports(store).by_index[1]
        assert(isinstance(update, Func))
        update(store)
        self.assertEqual(f(store), 5)
        self.assertEqual(f2(store), 5)

    def test_import_memory(self):
        store = Store()
        module = Module(store.engine, """
            (module
                (import "" "" (memory 1))
            )
        """)
        m = Memory(store, MemoryType(Limits(1, None)))
        Instance(store, module, [m])

    def test_import_table(self):
        store = Store()
        module = Module(store.engine, """
            (module
                (table (export "") 1 funcref)
            )
        """)
        table = Instance(store, module, []).exports(store).by_index[0]

        module = Module(store.engine, """
            (module
                (import "" "" (table 1 funcref))
            )
        """)
        Instance(store, module, [table])

    def test_invalid(self):
        store = Store()
        with self.assertRaises(AttributeError):
            Instance(store, 1, [])  # type: ignore
        with self.assertRaises(TypeError):
            Instance(store, Module(store.engine, '(module (import "" "" (func)))'), [1])  # type: ignore

        val = Func(store, FuncType([], []), lambda: None)
        module = Module(store.engine, '(module (import "" "" (func)))')
        Instance(store, module, [val])
        with self.assertRaises(WasmtimeError):
            Instance(store, module, [])
        with self.assertRaises(WasmtimeError):
            Instance(store, module, [val, val])

        module = Module(store.engine, '(module (import "" "" (global i32)))')
        with self.assertRaises(WasmtimeError):
            Instance(store, module, [val])

    def test_start_trap(self):
        store = Store()
        module = Module(store.engine, '(module (func unreachable) (start 0))')
        with self.assertRaises(Trap):
            Instance(store, module, [])



from typing import Generic, List, Union, cast, TypeVar


T = TypeVar('T')


class Slab(Generic[T]):
    list: List[Union[int, T]]
    next: int

    def __init__(self) -> None:
        self.list = []
        self.next = 0

    def allocate(self, val: T) -> int:
        idx = self.next

        if len(self.list) == idx:
            self.list.append(0)
            self.next += 1
        else:
            self.next = cast(int, self.list[idx])

        self.list[idx] = val
        return idx

    def get(self, idx: int) -> T:
        return cast(T, self.list[idx])

    def deallocate(self, idx: int) -> None:
        self.list[idx] = self.next
        self.next = idx



import unittest

from wasmtime import *


class TestTable(unittest.TestCase):
    def test_new(self):
        store = Store()
        module = Module(store.engine, """
            (module (table (export "") 1 funcref))
        """)
        table = Instance(store, module, []).exports(store).by_index[0]
        assert(isinstance(table, Table))
        assert(isinstance(table.type(store), TableType))
        self.assertEqual(table.type(store).limits, Limits(1, None))
        self.assertEqual(table.size(store), 1)

        ty = TableType(ValType.funcref(), Limits(1, 2))
        store = Store()
        func = Func(store, FuncType([], []), lambda: {})
        Table(store, ty, func)

    def test_grow(self):
        ty = TableType(ValType.funcref(), Limits(1, 2))
        store = Store()
        table = Table(store, ty, None)
        self.assertEqual(table.size(store), 1)

        # type errors
        with self.assertRaises(TypeError):
            table.grow(store, 'x', None)  # type: ignore
        with self.assertRaises(TypeError):
            table.grow(store, 2, 'x')

        # growth works
        table.grow(store, 1, None)
        self.assertEqual(table.size(store), 2)

        # can't grow beyond max
        with self.assertRaises(WasmtimeError):
            table.grow(store, 1, None)

    def test_get(self):
        ty = TableType(ValType.funcref(), Limits(1, 2))
        store = Store()
        table = Table(store, ty, None)
        self.assertEqual(table.get(store, 0), None)
        self.assertEqual(table.get(store, 1), None)

        called = {}
        called['hit'] = False

        def set_called():
            called['hit'] = True
        func = Func(store, FuncType([], []), set_called)
        table.grow(store, 1, func)
        assert(not called['hit'])
        f = table.get(store, 1)
        assert(isinstance(f, Func))
        f(store)
        assert(called['hit'])

    def test_set(self):
        ty = TableType(ValType.funcref(), Limits(1, 2))
        store = Store()
        table = Table(store, ty, None)
        func = Func(store, FuncType([], []), lambda: {})
        table.set(store, 0, func)
        with self.assertRaises(WasmtimeError):
            table.set(store, 1, func)



import ctypes

from . import _ffi as ffi
from wasmtime import WasmtimeError
import typing


def wat2wasm(wat: typing.Union[str, bytes]) -> bytearray:
    """
    Converts the [WebAssembly Text format][wat] to the binary format.

    This function is intended to be a convenience function for local
    development and you likely don't want to use it extensively in production.
    It's much faster to parse and compile the binary format than it is to
    process the text format.

    Takes a `str` as input, raises an error if it fails to parse, and returns
    a `bytes` if conversion/parsing was successful.

    >>> wat2wasm('(module)')
    bytearray(b'\\x00asm\\x01\\x00\\x00\\x00')

    [wat]: https://webassembly.github.io/spec/core/text/index.html
    """

    if isinstance(wat, str):
        wat = wat.encode('utf8')
    wat_buffer = ctypes.create_string_buffer(wat)
    wasm = ffi.wasm_byte_vec_t()
    error = ffi.wasmtime_wat2wasm(wat_buffer, len(wat), ctypes.byref(wasm))
    if error:
        raise WasmtimeError._from_ptr(error)
    else:
        ret = ffi.to_bytes(wasm)
        ffi.wasm_byte_vec_delete(ctypes.byref(wasm))
        return ret

def _to_wasm(wasm: typing.Union[str, bytes, bytearray]) -> typing.Union[bytes, bytearray]:
    # If this looks like a string, parse it as the text format. Note that
    # in python 2 strings and bytes are basically the same, so we skip this
    # if the first byte in the string is 0, meaning this is actually a wasm
    # module.
    if isinstance(wasm, str) and len(wasm) > 0 and ord(wasm[0]) != 0:
        wasm = wat2wasm(wasm)
    if isinstance(wasm, bytes) and len(wasm) > 0 and wasm[0] != 0:
        wasm = wat2wasm(wasm)

    if not isinstance(wasm, (bytes, bytearray)):
        raise TypeError("expected wasm bytes")
    return wasm



from . import _ffi as ffi
from ._wat2wasm import _to_wasm
import ctypes
from wasmtime import Engine, wat2wasm, ImportType, ExportType, WasmtimeError, Managed
import typing
from os import PathLike


class Module(Managed["ctypes._Pointer[ffi.wasmtime_module_t]"]):

    @classmethod
    def from_file(cls, engine: Engine, path: typing.Union[str, bytes, PathLike]) -> "Module":
        """
        Compiles and creates a new `Module` by reading the file at `path` and
        then delegating to the `Module` constructor.
        """

        with open(path, "rb") as f:
            contents = f.read()
        return cls(engine, contents)

    def __init__(self, engine: Engine, wasm: typing.Union[str, bytes]):
        if not isinstance(engine, Engine):
            raise TypeError("expected an Engine")

        wasm = _to_wasm(wasm)

        # TODO: can the copy be avoided here? I can't for the life of me
        # figure this out.
        binary = (ctypes.c_uint8 * len(wasm)).from_buffer_copy(wasm)
        ptr = ctypes.POINTER(ffi.wasmtime_module_t)()
        error = ffi.wasmtime_module_new(engine.ptr(), binary, len(wasm), ctypes.byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        self._set_ptr(ptr)

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_module_t]") -> None:
        ffi.wasmtime_module_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_module_t]") -> "Module":
        if not isinstance(ptr, ctypes.POINTER(ffi.wasmtime_module_t)):
            raise TypeError("wrong pointer type")
        ty: "Module" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @classmethod
    def deserialize(cls, engine: Engine, encoded: typing.Union[bytes, bytearray]) -> 'Module':
        """
        Deserializes bytes previously created by `Module.serialize`.

        This constructor for `Module` will deserialize bytes previously created
        by a serialized module. This will only succeed if the bytes were
        previously created by the same version of `wasmtime` as well as the
        same configuration within `Engine`.
        """

        if not isinstance(encoded, (bytes, bytearray)):
            raise TypeError("expected bytes")

        ptr = ctypes.POINTER(ffi.wasmtime_module_t)()

        # TODO: can the copy be avoided here? I can't for the life of me
        # figure this out.
        error = ffi.wasmtime_module_deserialize(
            engine.ptr(),
            (ctypes.c_uint8 * len(encoded)).from_buffer_copy(encoded),
            len(encoded),
            ctypes.byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        return cls._from_ptr(ptr)

    @classmethod
    def deserialize_file(cls, engine: Engine, path: str) -> 'Module':
        """
        Deserializes bytes previously created by `Module.serialize` that are
        stored in a file on the filesystem.

        Otherwise this function is the same as `Module.deserialize`.
        """

        ptr = ctypes.POINTER(ffi.wasmtime_module_t)()
        path_bytes = path.encode('utf-8')
        error = ffi.wasmtime_module_deserialize_file(
            engine.ptr(),
            path_bytes,
            ctypes.byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        return cls._from_ptr(ptr)

    @classmethod
    def validate(cls, engine: Engine, wasm: typing.Union[bytes, bytearray]) -> None:
        """
        Validates whether the list of bytes `wasm` provided is a valid
        WebAssembly binary given the configuration in `store`

        Raises a `WasmtimeError` if the wasm isn't valid.
        """

        if not isinstance(wasm, (bytes, bytearray)):
            raise TypeError("expected wasm bytes")

        # TODO: can the copy be avoided here? I can't for the life of me
        # figure this out.
        buf = (ctypes.c_uint8 * len(wasm)).from_buffer_copy(wasm)
        error = ffi.wasmtime_module_validate(engine.ptr(), buf, len(wasm))

        if error:
            raise WasmtimeError._from_ptr(error)

    @property
    def imports(self) -> typing.List[ImportType]:
        """
        Returns the types of imports that this module has
        """

        imports = ImportTypeList()
        ffi.wasmtime_module_imports(self.ptr(), ctypes.byref(imports.vec))
        ret = []
        for i in range(0, imports.vec.size):
            ret.append(ImportType._from_ptr(imports.vec.data[i], imports))
        return ret

    @property
    def exports(self) -> typing.List[ExportType]:
        """
        Returns the types of the exports that this module has
        """

        exports = ExportTypeList()
        ffi.wasmtime_module_exports(self.ptr(), ctypes.byref(exports.vec))
        ret = []
        for i in range(0, exports.vec.size):
            ret.append(ExportType._from_ptr(exports.vec.data[i], exports))
        return ret

    def serialize(self) -> bytearray:
        """
        Serializes this module to a binary representation.

        This method will serialize this module to an in-memory byte array which
        can be cached and later passed to `Module.deserialize` to recreate this
        module.
        """
        raw = ffi.wasm_byte_vec_t()
        err = ffi.wasmtime_module_serialize(self.ptr(), ctypes.byref(raw))
        if err:
            raise WasmtimeError._from_ptr(err)
        ret = ffi.to_bytes(raw)
        ffi.wasm_byte_vec_delete(ctypes.byref(raw))
        return ret


class ImportTypeList:
    def __init__(self) -> None:
        self.vec = ffi.wasm_importtype_vec_t(0, None)

    def __del__(self) -> None:
        ffi.wasm_importtype_vec_delete(ctypes.byref(self.vec))


class ExportTypeList:
    def __init__(self) -> None:
        self.vec = ffi.wasm_exporttype_vec_t(0, None)

    def __del__(self) -> None:
        ffi.wasm_exporttype_vec_delete(ctypes.byref(self.vec))



import unittest

from wasmtime import Engine, WasmtimeError, Store
from wasmtime.component import Linker, Component, Instance


class TestInstance(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(WasmtimeError):
            Instance()

    def test_smoke(self):
        engine = Engine()
        linker = Linker(engine)
        component = Component(engine, '(component)')
        store = Store(engine)

        instance = linker.instantiate(store, component)
        self.assertIsNotNone(instance)
        self.assertIsNone(instance.get_export_index(store, 'hello'))

        component = Component(engine, """
            (component
                (core module (export "foo"))
            )
        """)
        instance = linker.instantiate(store, component)
        self.assertIsNotNone(instance.get_export_index(store, 'foo'))

        component = Component(engine, """
            (component
                (core module $a)
                (instance (export "x")
                    (export "m" (core module $a))
                )
            )
        """)
        instance = linker.instantiate(store, component)
        e1 = instance.get_export_index(store, 'x')
        e2 = component.get_export_index('x')
        assert(e1 is not None)
        assert(e2 is not None)
        self.assertIsNotNone(instance.get_export_index(store, 'm', instance = e1))
        self.assertIsNotNone(instance.get_export_index(store, 'm', instance = e2))

        self.assertIsNone(instance.get_func(store, e1))
        self.assertIsNone(instance.get_func(store, e2))

    def test_get_func(self):
        engine = Engine()
        linker = Linker(engine)
        store = Store(engine)

        component = Component(engine, """
            (component
                (core module $a
                    (func (export "foo"))
                )
                (core instance $a (instantiate $a))
                (func (export "a") (canon lift (core func $a "foo")))
            )
        """)
        instance = linker.instantiate(store, component)
        index = instance.get_export_index(store, 'a')
        assert(index is not None)
        func = instance.get_func(store, index)
        self.assertIsNotNone(func)

        self.assertIsNotNone(instance.get_func(store, 'a'))
        self.assertIsNone(instance.get_func(store, 'b'))



from contextlib import contextmanager
import ctypes
from .. import _ffi as ffi, StoreContext, WasmtimeError
from typing import Optional, Callable


LAST_EXCEPTION: Optional[Exception] = None


def catch_exceptions(store_raw: 'ctypes._Pointer[ffi.wasmtime_context_t]', func: Callable[[StoreContext], None]) -> ctypes.c_size_t:
    store = StoreContext(store_raw)
    exception = None
    try:
        func(store)
    except WasmtimeError as e:
        exception = e
    except Exception as e:
        global LAST_EXCEPTION
        LAST_EXCEPTION = e
        exception = WasmtimeError("python exception")
    finally:
        store._invalidate()

    if exception:
        return ctypes.cast(exception._consume(), ctypes.c_void_p).value # type: ignore
    return 0 # type: ignore


def enter_wasm(func: Callable[[], 'ctypes._Pointer[ffi.wasmtime_error_t]']) -> None:
    ptr = func()
    if ptr:
        error = WasmtimeError._from_ptr(ptr)
        maybe_raise_last_exn()
        raise error


def maybe_raise_last_exn() -> None:
    global LAST_EXCEPTION
    if LAST_EXCEPTION is None:
        return
    exn = LAST_EXCEPTION
    LAST_EXCEPTION = None
    raise exn



import unittest

from wasmtime import FuncType, ValType


class TestTypes(unittest.TestCase):
    def test_valtypes(self):
        ValType.i32()
        ValType.i64()
        ValType.f32()
        ValType.f64()

    def test_new(self):
        FuncType([], [])
        FuncType([ValType.i32()], [ValType.i64()])



import unittest

from wasmtime import *


class TestLinker(unittest.TestCase):
    def test_define(self):
        store = Store()
        linker = Linker(store.engine)
        linker.allow_shadowing = False

        func = Func(store, FuncType([], []), lambda: None)
        linker.define(store, "", "a", func)

        g = Global(store, GlobalType(ValType.i32(), False), Val.i32(0))
        linker.define(store, "", "c", g)

        mem = Memory(store, MemoryType(Limits(1, None)))
        linker.define(store, "", "e", mem)

        module = Module(store.engine, """
            (module (table (export "") 1 funcref))
        """)
        table = Instance(store, module, []).exports(store).by_index[0]
        linker.define(store, "", "g", table)

        with self.assertRaises(WasmtimeError):
            linker.define(store, "", "a", func)
        linker.allow_shadowing = True
        linker.define(store, "", "a", func)

        with self.assertRaises(TypeError):
            linker.define(store, "", "", 2)  # type: ignore
        with self.assertRaises(AttributeError):
            linker.define(store, 2, "", func)  # type: ignore
        with self.assertRaises(AttributeError):
            linker.define(store, "", 2, func)  # type: ignore

    def test_define_instance(self):
        store = Store()
        linker = Linker(store.engine)
        with self.assertRaises(TypeError):
            linker.define_instance("x", 2)  # type: ignore

        module = Module(store.engine, "(module)")
        linker.define_instance(store, "a", Instance(store, module, []))

        module = Module(store.engine, "(module (func (export \"foo\")))")
        instance = Instance(store, module, [])
        linker.define_instance(store, "b", instance)
        with self.assertRaises(WasmtimeError):
            linker.define_instance(store, "b", instance)
        linker.allow_shadowing = True
        linker.define_instance(store, "b", instance)

    def test_define_wasi(self):
        linker = Linker(Engine())
        linker.define_wasi()

    def test_instantiate(self):
        store = Store()
        linker = Linker(store.engine)

        module = Module(store.engine, "(module (func (export \"foo\")))")
        instance = Instance(store, module, [])
        linker.define_instance(store, "x", instance)

        func = Func(store, FuncType([], []), lambda: None)
        linker.define(store, "y", "z", func)

        module = Module(store.engine, """
            (module
                (import "x" "foo" (func))
                (import "y" "z" (func))
            )
        """)
        linker.instantiate(store, module)

        module = Module(store.engine, """
            (module
                (import "x" "foo" (func))
                (import "y" "z" (global i32))
            )
        """)
        with self.assertRaises(WasmtimeError):
            linker.instantiate(store, module)

        module = Module(store.engine, """
            (module
                (func unreachable)
                (start 0)
            )
        """)
        with self.assertRaises(Trap):
            linker.instantiate(store, module)

        module = Module(store.engine, "(module)")
        linker.instantiate(store, module)

    def test_errors(self):
        linker = Linker(Engine())
        with self.assertRaises(AttributeError):
            Linker(2)  # type: ignore
        with self.assertRaises(AttributeError):
            linker.instantiate(Store(), 3)  # type: ignore

    def test_module(self):
        store = Store()
        linker = Linker(store.engine)
        module = Module(store.engine, """
            (module
                (func (export "f"))
            )
        """)
        linker.define_module(store, "foo", module)
        module = Module(store.engine, """
            (module
                (import "foo" "f" (func))
            )
        """)
        linker.instantiate(store, module)

    def test_get_default(self):
        store = Store()
        linker = Linker(store.engine)
        linker.get_default(store, "foo")(store)

    def test_get_one_by_name(self):
        store = Store()
        linker = Linker(store.engine)
        with self.assertRaises(WasmtimeError):
            linker.get(store, "foo", "bar")
        module = Module(store.engine, """
            (module
                (func (export "f"))
            )
        """)
        linker.define_module(store, "foo", module)
        assert(isinstance(linker.get(store, "foo", "f"), Func))

    def test_define_func(self):
        engine = Engine()
        linker = Linker(engine)
        called = {}
        called['hits'] = 0

        def call():
            called['hits'] += 1

        linker.define_func('a', 'b', FuncType([], []), call)
        module = Module(engine, """
            (module
                (import "a" "b" (func))
                (start 0)
            )
        """)
        assert(called['hits'] == 0)
        linker.instantiate(Store(engine), module)
        assert(called['hits'] == 1)
        linker.instantiate(Store(engine), module)
        assert(called['hits'] == 2)



import ctypes
import errno
from ctypes import POINTER, c_char, c_char_p, cast, CFUNCTYPE, c_void_p
from enum import Enum
from os import PathLike
from typing import Iterable, List, Union, Callable

from wasmtime import Managed, WasmtimeError

from . import _ffi as ffi
from ._config import setter_property
from ._slab import Slab


def _encode_path(path: Union[str, bytes, PathLike]) -> bytes:
    if isinstance(path, (bytes, str)):
        path2 = path
    else:
        path2 = path.__fspath__()
    if isinstance(path2, bytes):
        return path2
    return path2.encode('utf8')

class DirPerms(Enum):
    READ_ONLY = ffi.wasi_dir_perms_flags.WASMTIME_WASI_DIR_PERMS_READ.value
    WRITE_ONLY = ffi.wasi_dir_perms_flags.WASMTIME_WASI_DIR_PERMS_WRITE.value
    READ_WRITE = ffi.wasi_dir_perms_flags.WASMTIME_WASI_DIR_PERMS_READ.value | ffi.wasi_dir_perms_flags.WASMTIME_WASI_DIR_PERMS_WRITE.value

class FilePerms(Enum):
    READ_ONLY = ffi.wasi_file_perms_flags.WASMTIME_WASI_FILE_PERMS_READ.value
    WRITE_ONLY = ffi.wasi_file_perms_flags.WASMTIME_WASI_FILE_PERMS_WRITE.value
    READ_WRITE = ffi.wasi_file_perms_flags.WASMTIME_WASI_FILE_PERMS_READ.value | ffi.wasi_file_perms_flags.WASMTIME_WASI_FILE_PERMS_WRITE.value


CustomOutput = Callable[[bytes], Union[int, None]]
CUSTOM_OUTPUTS: Slab[CustomOutput] = Slab()


class WasiConfig(Managed["ctypes._Pointer[ffi.wasi_config_t]"]):

    def __init__(self) -> None:
        self._set_ptr(ffi.wasi_config_new())

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasi_config_t]") -> None:
        ffi.wasi_config_delete(ptr)

    @setter_property
    def argv(self, argv: List[str]) -> None:
        """
        Explicitly configure the `argv` for this WASI configuration
        """
        ptrs = to_char_array(argv)
        if not ffi.wasi_config_set_argv(self.ptr(), len(argv), ptrs):
            raise WasmtimeError("failed to configure argv")

    def inherit_argv(self) -> None:
        ffi.wasi_config_inherit_argv(self.ptr())

    @setter_property
    def env(self, pairs: Iterable[Iterable]) -> None:
        """
        Configure environment variables to be returned for this WASI
        configuration.

        The `pairs` provided must be an iterable list of key/value pairs of
        environment variables.
        """
        names = []
        values = []
        for name, value in pairs:
            names.append(name)
            values.append(value)
        name_ptrs = to_char_array(names)
        value_ptrs = to_char_array(values)
        if not ffi.wasi_config_set_env(self.ptr(), len(names), name_ptrs, value_ptrs):
            raise WasmtimeError("failed to configure environment")

    def inherit_env(self) -> None:
        """
        Configures the environment variables available within WASI to be those
        in this own process's environment. All environment variables are
        inherited.
        """
        ffi.wasi_config_inherit_env(self.ptr())

    @setter_property
    def stdin_file(self, path: Union[str, bytes, PathLike]) -> None:
        """
        Configures a file to be used as the stdin stream of this WASI
        configuration.

        Reads of the stdin stream will read the path specified.

        The file must already exist on the filesystem. If it cannot be
        opened then `WasmtimeError` is raised.
        """

        res = ffi.wasi_config_set_stdin_file(
            self.ptr(), c_char_p(_encode_path(path)))
        if not res:
            raise WasmtimeError("failed to set stdin file")

    def inherit_stdin(self) -> None:
        """
        Configures this own process's stdin to be used as the WASI program's
        stdin.

        Reads of the stdin stream will read this process's stdin.
        """
        ffi.wasi_config_inherit_stdin(self.ptr())

    @setter_property
    def stdout_file(self, path: str) -> None:
        """
        Configures a file to be used as the stdout stream of this WASI
        configuration.

        Writes to stdout will be written to the file specified.

        The file specified will be created if it doesn't exist, or truncated if
        it already exists. It must be available to open for writing. If it
        cannot be opened for writing then `WasmtimeError` is raised.
        """
        res = ffi.wasi_config_set_stdout_file(
            self.ptr(), c_char_p(_encode_path(path)))
        if not res:
            raise WasmtimeError("failed to set stdout file")

    @setter_property
    def stdout_custom(self, callback: CustomOutput) -> None:
        """
        Sets a custom `callback` that is invoked whenever stdout is written to.
        """
        ffi.wasi_config_set_stdout_custom(
            self.ptr(), custom_call,
            CUSTOM_OUTPUTS.allocate(callback), custom_finalize)

    def inherit_stdout(self) -> None:
        """
        Configures this own process's stdout to be used as the WASI program's
        stdout.

        Writes to stdout stream will write to this process's stdout.
        """
        ffi.wasi_config_inherit_stdout(self.ptr())

    @setter_property
    def stderr_file(self, path: str) -> None:
        """
        Configures a file to be used as the stderr stream of this WASI
        configuration.

        Writes to stderr will be written to the file specified.

        The file specified will be created if it doesn't exist, or truncated if
        it already exists. It must be available to open for writing. If it
        cannot be opened for writing then `WasmtimeError` is raised.
        """
        res = ffi.wasi_config_set_stderr_file(
            self.ptr(), c_char_p(_encode_path(path)))
        if not res:
            raise WasmtimeError("failed to set stderr file")

    @setter_property
    def stderr_custom(self, callback: CustomOutput) -> None:
        """
        Sets a custom `callback` that is invoked whenever stderr is written to.
        """
        ffi.wasi_config_set_stderr_custom(
            self.ptr(), custom_call,
            CUSTOM_OUTPUTS.allocate(callback), custom_finalize)

    def inherit_stderr(self) -> None:
        """
        Configures this own process's stderr to be used as the WASI program's
        stderr.

        Writes to stderr stream will write to this process's stderr.
        """
        ffi.wasi_config_inherit_stderr(self.ptr())

    def preopen_dir(self, path: str, guest_path: str, dir_perms: DirPerms = DirPerms.READ_WRITE, file_perms: FilePerms = FilePerms.READ_WRITE) -> None:
        """
        Allows the WASI program to access the directory at `path` using the
        path `guest_path` within the WASI program.

        `dir_perms` specifies the permissions that wasm will have to operate on
        `guest_path`. This can be used, for example, to provide readonly access to a
        directory.

        `file_perms` specifies the maximum set of permissions that can be used for
        any file in this directory.
        """
        path_ptr = c_char_p(path.encode('utf-8'))
        guest_path_ptr = c_char_p(guest_path.encode('utf-8'))
        if not ffi.wasi_config_preopen_dir(self.ptr(), path_ptr, guest_path_ptr, dir_perms.value, file_perms.value):
            raise WasmtimeError('failed to add preopen dir')


def to_char_array(strings: List[str]) -> "ctypes._Pointer[ctypes._Pointer[c_char]]":
    ptrs = (c_char_p * len(strings))()
    for i, s in enumerate(strings):
        ptrs[i] = c_char_p(s.encode('utf-8'))
    return cast(ptrs, POINTER(POINTER(c_char)))


@CFUNCTYPE(ctypes.c_ssize_t, c_void_p, POINTER(ctypes.c_ubyte), ctypes.c_size_t)
def custom_call(idx, ptr, size): # type: ignore
    try:
        ty = ctypes.c_uint8 * size
        arg = bytes(ty.from_address(ctypes.addressof(ptr.contents)))
        ret = CUSTOM_OUTPUTS.get(idx or 0)(arg)
        if ret is None:
            return size
        return ret
    except Exception as e:
        print('failed custom output, required to catch exception:', e)
        return -errno.EIO


@CFUNCTYPE(None, c_void_p)
def custom_finalize(idx): # type: ignore
    if CUSTOM_OUTPUTS:
        CUSTOM_OUTPUTS.deallocate(idx or 0)
    return None



import unittest
from dataclasses import dataclass
from wasmtime import Store, WasmtimeError, Engine
from wasmtime.component import *
from typing import Any, List

def list_component(ty_wit: str):
    return f"""
        (component
            (import "x" (func $x (param "x" {ty_wit}) (result {ty_wit})))
            (core module $libc
                (memory (export "mem") 1)
                (global $base (mut i32) (i32.const 100))
                (func (export "realloc") (param i32 i32 i32 i32) (result i32)
                    (local $ret i32)
                    local.get 0
                    if unreachable end
                    local.get 1
                    if unreachable end

                    (local.set $ret (global.get $base))
                    (global.set $base
                        (i32.add
                            (global.get $base)
                            (local.get 2)))
                    local.get $ret)
            )
            (core instance $libc (instantiate $libc))
            (core module $a
                (import "" "x" (func (param i32 i32 i32)))
                (func (export "x") (param i32 i32)  (result i32)
                    local.get 0
                    local.get 1
                    i32.const 0
                    call 0
                    i32.const 0)
            )
            (core func $x (canon lower (func $x)
                (memory $libc "mem") (realloc (func $libc "realloc"))))
            (core instance $a (instantiate $a
                (with "" (instance
                    (export "x" (func $x))
                ))
            ))
            (func (export "x") (param "x" {ty_wit}) (result {ty_wit})
                (canon lift (core func $a "x") (memory $libc "mem")
                (realloc (func $libc "realloc"))))
        )
    """

def retptr_component(ty_wit: str, ty_wasm: str):
    return f"""
        (component
            (type $t' {ty_wit})
            (import "t" (type $t (eq $t')))
            (import "x" (func $x (param "x" $t) (result $t)))
            (core module $libc (memory (export "mem") 1))
            (core instance $libc (instantiate $libc))
            (core module $a
                (import "" "x" (func (param {ty_wasm} i32)))
                (func (export "x") (param {ty_wasm})  (result i32)
                    local.get 0
                    local.get 1
                    i32.const 0
                    call 0
                    i32.const 0)
            )
            (core func $x (canon lower (func $x) (memory $libc "mem")))
            (core instance $a (instantiate $a
                (with "" (instance
                    (export "x" (func $x))
                ))
            ))
            (func (export "x") (param "x" $t) (result $t)
                (canon lift (core func $a "x") (memory $libc "mem")))
        )
    """

class TestFunc(unittest.TestCase):
    def roundtrip(self, wat: str, values: List[Any], bad = TypeError) -> None:
        engine = Engine()
        store = Store(engine)
        component = Component(engine, wat)
        value_being_tested = None

        def roundtrip(_store, val):
            self.assertEqual(value_being_tested, val)
            return val

        linker = Linker(engine)
        with linker.root() as l:
            l.add_func('x', roundtrip)
        instance = linker.instantiate(store, component)
        f = instance.get_func(store, 'x')
        assert(f is not None)
        for val in values:
            value_being_tested = val
            ret = f(store, val)
            self.assertEqual(ret, val)
            f.post_return(store)

        class Bad:
            pass
        with self.assertRaises(bad):
            f(store, Bad())

    def roundtrip_simple(self, ty_wit: str, ty_wasm: str, values: List[Any]) -> None:
        wat = f"""
            (component
                (type $t' {ty_wit})
                (import "t" (type $t (eq $t')))
                (import "x" (func $x (param "x" $t) (result $t)))
                (core module $a
                    (import "" "x" (func (param {ty_wasm}) (result {ty_wasm})))
                    (func (export "x") (param {ty_wasm}) (result {ty_wasm})
                        local.get 0
                        call 0)
                )
                (core func $x (canon lower (func $x)))
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "x" (func $x))
                    ))
                ))
                (func (export "x") (param "x" $t) (result $t)
                    (canon lift (core func $a "x")))
            )
        """
        self.roundtrip(wat, values)

    def test_init(self):
        with self.assertRaises(WasmtimeError):
            Func()

    def test_type_reflection(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, """
            (component
                (core module $a
                    (func (export "a"))
                    (func (export "b") (param i32) (result i32) unreachable)
                )
                (core instance $a (instantiate $a))
                (func (export "a") (canon lift (core func $a "a")))
                (func (export "b") (param "x" u32) (result u32)
                    (canon lift (core func $a "b")))
            )
        """)
        instance = Linker(engine).instantiate(store, component)
        ai = instance.get_export_index(store, 'a')
        bi = instance.get_export_index(store, 'b')
        assert(ai is not None)
        assert(bi is not None)
        a = instance.get_func(store, ai)
        b = instance.get_func(store, bi)
        assert(a is not None)
        assert(b is not None)
        self.assertEqual(a.type(store).params, [])
        self.assertIsNone(a.type(store).result)
        self.assertEqual(b.type(store).params, [('x', U32())])
        self.assertEqual(b.type(store).result, U32())

    def test_call(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, """
            (component
                (core module $a
                    (func (export "a"))
                    (func (export "b") (param i32)
                        local.get 0
                        i32.const 100
                        i32.ne
                        if unreachable end)
                    (func (export "c") (result i32) (i32.const 101))
                )
                (core instance $a (instantiate $a))
                (func (export "a") (canon lift (core func $a "a")))
                (func (export "b") (param "x" u32) (canon lift (core func $a "b")))
                (func (export "c") (result u32) (canon lift (core func $a "c")))
            )
        """)
        instance = Linker(engine).instantiate(store, component)
        a = instance.get_func(store, 'a')
        assert(a is not None)
        ret = a(store)
        self.assertEqual(ret, None)
        a.post_return(store)

        b = instance.get_func(store, 'b')
        assert(b is not None)
        ret = b(store, 100)
        self.assertEqual(ret, None)
        b.post_return(store)

        c = instance.get_func(store, 'c')
        assert(c is not None)
        ret = c(store)
        self.assertEqual(ret, 101)
        c.post_return(store)

        with self.assertRaises(TypeError):
            b(store)

    def test_roundtrip_empty(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, f"""
            (component
                (import "x" (func $x))
                (core module $a
                    (import "" "x" (func))
                    (func (export "x") call 0)
                )
                (core func $x (canon lower (func $x)))
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "x" (func $x))
                    ))
                ))
                (func (export "x") (canon lift (core func $a "x")))
            )
        """)

        linker = Linker(engine)
        with linker.root() as l:
            l.add_func('x', lambda _store: None)
        instance = linker.instantiate(store, component)
        f = instance.get_func(store, 'x')
        assert(f is not None)

        self.assertIsNone(f(store))
        f.post_return(store)

    def test_roundtrip_primitive(self):
        self.roundtrip_simple('bool', 'i32', [True, False])
        self.roundtrip_simple('u8', 'i32', [0, 1, 42, 255])
        self.roundtrip_simple('u16', 'i32', [0, 1, 42, 65535])
        self.roundtrip_simple('u32', 'i32', [0, 1, 42, 4294967295])
        self.roundtrip_simple('u64', 'i64', [0, 1, 42, 18446744073709551615])
        self.roundtrip_simple('s8', 'i32', [0, 1, -1, 42, -42, 127, -128])
        self.roundtrip_simple('s16', 'i32', [0, 1, -1, 42, -42, 32767, -32768])
        self.roundtrip_simple('s32', 'i32', [0, 1, -1, 42, -42, 2147483647, -2147483648])
        self.roundtrip_simple('s64', 'i64', [0, 1, -1, 42, -42, 9223372036854775807, -9223372036854775808])
        self.roundtrip_simple('f32', 'f32', [0.0, 1.0, -1.0])
        self.roundtrip_simple('f64', 'f64', [0.0, 1.0, -1.0])
        self.roundtrip_simple('char', 'i32', ['a', 'b'])

    def test_resources(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, f"""
            (component
                (import "t" (type $t (sub resource)))
                (import "mk" (func $mk (result (own $t))))
                (import "borrow" (func $borrow (param "t" (borrow $t))))
                (import "own" (func $own (param "t" (own $t))))
                (core module $a
                    (import "" "mk" (func $mk (result i32)))
                    (import "" "borrow" (func $borrow (param i32)))
                    (import "" "own" (func $own (param i32)))
                    (import "" "drop" (func $drop (param i32)))

                    (func (export "mk") (result i32) call $mk)
                    (func (export "borrow") (param i32)
                        (call $borrow (local.get 0))
                        (call $drop (local.get 0)))
                    (func (export "own") (param i32) local.get 0 call $own)
                )
                (core func $mk (canon lower (func $mk)))
                (core func $borrow (canon lower (func $borrow)))
                (core func $own (canon lower (func $own)))
                (core func $drop (canon resource.drop $t))
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "mk" (func $mk))
                        (export "borrow" (func $borrow))
                        (export "own" (func $own))
                        (export "drop" (func $drop))
                    ))
                ))
                (func (export "mk") (result (own $t))
                    (canon lift (core func $a "mk")))
                (func (export "borrow") (param "x" (borrow $t))
                    (canon lift (core func $a "borrow")))
                (func (export "own") (param "x" (own $t))
                    (canon lift (core func $a "own")))
            )
        """)

        ty = ResourceType.host(2)

        def mk(_store):
            return ResourceHost.own(1, 2)

        def borrow(_store, b):
            self.assertFalse(b.owned)
            handle = b.to_host(store)
            self.assertEqual(handle.type, 2)

        def own(_store, b):
            self.assertTrue(b.owned)

        linker = Linker(engine)
        with linker.root() as l:
            l.add_resource('t', ty, lambda _store, _rep: None)
            l.add_func('mk', mk)
            l.add_func('borrow', borrow)
            l.add_func('own', own)
        instance = linker.instantiate(store, component)
        f_mk = instance.get_func(store, 'mk')
        f_borrow = instance.get_func(store, 'borrow')
        f_own = instance.get_func(store, 'own')
        assert(f_mk is not None)
        assert(f_borrow is not None)
        assert(f_own is not None)

        r1 = f_mk(store)
        self.assertIsInstance(r1, ResourceAny)
        f_mk.post_return(store)
        f_borrow(store, r1)
        f_borrow.post_return(store)
        f_borrow(store, ResourceHost.own(1, 2))
        f_borrow.post_return(store)
        f_own(store, r1)
        f_own.post_return(store)
        f_own(store, ResourceHost.own(1, 2))
        f_own.post_return(store)

        with self.assertRaises(TypeError):
            f_borrow(store, 1)

        with self.assertRaises(TypeError):
            f_own(store, 1)


    def test_enum(self):
        self.roundtrip_simple('(enum "a" "b" "c")', 'i32', ['a', 'b', 'c'])
        with self.assertRaises(WasmtimeError):
            self.roundtrip_simple('(enum "a" "b" "c")', 'i32', ['d'])


    def test_flags(self):
        vals = [{'a'}, {'b'}, {'c'}, {'a', 'b'}, {'a', 'c'}, {'b', 'c'}]
        self.roundtrip_simple('(flags "a" "b" "c")', 'i32', vals)
        with self.assertRaises(WasmtimeError):
            self.roundtrip_simple('(flags "a" "b" "c")', 'i32', [{'d'}])


    def test_record(self):
        ty_wit = '(record (field "a" u32) (field "b" bool))'
        ty_wasm = 'i32 i32'
        wat = retptr_component(ty_wit, ty_wasm)
        @dataclass
        class Record:
            a: int
            b: bool
        values = [
                Record(0, False),
                Record(1, True),
                Record(42, False),
                Record(65535, True)
        ]
        self.roundtrip(wat, values, bad = AttributeError)

        @dataclass
        class RecordBad:
            a: int
        with self.assertRaises(AttributeError):
            self.roundtrip(wat, [RecordBad(0)])

    def test_tuple(self):
        ty_wit = '(tuple u32 bool)'
        ty_wasm = 'i32 i32'
        wat = retptr_component(ty_wit, ty_wasm)
        values = [(0, False), (1, True), (42, False), (65535, True)]
        self.roundtrip(wat, values)

        with self.assertRaises(TypeError):
            self.roundtrip(wat, [(0,)])

        with self.assertRaises(TypeError):
            self.roundtrip(wat, [(0, 'x')])

    def test_string(self):
        wat = list_component('string')
        self.roundtrip(wat, ['', 'a', 'hello, world!'])

    def test_list(self):
        wat = list_component('(list u8)')
        self.roundtrip(wat, [b'', b'a', b'hello, world!'])
        wat = list_component('(list u32)')
        self.roundtrip(wat, [[], [1], [1, 2, 3, 4, 5]])
        with self.assertRaises(TypeError):
            self.roundtrip(wat, [[1, 2, 3, 'x']])

    def test_variant(self):
        self.roundtrip_simple('(variant (case "a") (case "b"))',
                              'i32',
                              [Variant('a'), Variant('b')])

        with self.assertRaises(ValueError):
            self.roundtrip_simple('(variant (case "a") (case "b"))',
                                  'i32',
                                  [Variant('c')])

        wat = retptr_component('(variant (case "a" u32) (case "b" u64))', 'i32 i64')
        self.roundtrip(wat, [Variant('a', 1), Variant('b', 2)])

        wat = retptr_component('(variant (case "a" u32) (case "b" f32))', 'i32 i32')
        self.roundtrip(wat, [1, 2.], bad=ValueError)
        wat = retptr_component('(variant (case "a") (case "b" f32))', 'i32 f32')
        self.roundtrip(wat, [None, 2.], bad=ValueError)


    def test_option(self):
        wat = retptr_component('(option u32)', 'i32 i32')
        self.roundtrip(wat, [None, 3, 0], bad=ValueError)


    def test_result(self):
        wat = retptr_component('(result u32 (error f32))', 'i32 i32')
        self.roundtrip(wat, [3, 1.], bad=ValueError)

        wat = retptr_component('(result u32)', 'i32 i32')
        self.roundtrip(wat, [3, None], bad=ValueError)

        wat = retptr_component('(result (error f32))', 'i32 f32')
        self.roundtrip(wat, [3., None], bad=ValueError)

        self.roundtrip_simple('(result)', 'i32', [Variant('ok'), Variant('err')])


    # TODO: roundtrip future
    # TODO: roundtrip stream
    # TODO: roundtrip error-context
    # TODO: typechecking in variants-of-python-types (e.g. `add_classes`)



from contextlib import contextmanager
from ctypes import POINTER, byref, CFUNCTYPE, c_void_p, cast
import ctypes
from wasmtime import Store, FuncType, Val, Trap, WasmtimeError
from . import _ffi as ffi
from ._extern import wrap_extern
from typing import Callable, Optional, Generic, TypeVar, List, Union, Tuple, cast as cast_type, Sequence, Any
from ._exportable import AsExtern
from ._store import Storelike
from ._slab import Slab


FUNCTIONS: "Slab[Tuple]"
LAST_EXCEPTION: Optional[Exception] = None


class Func:
    _func: ffi.wasmtime_func_t

    def __init__(self, store: Storelike, ty: FuncType, func: Callable, access_caller: bool = False):
        """
        Creates a new func in `store` with the given `ty` which calls the closure
        given

        The `func` is called with the parameters natively and they'll have native
        Python values rather than being wrapped in `Val`. If `access_caller` is
        set to `True` then the first argument given to `func` is an instance of
        type `Caller` below.
        """

        if not isinstance(store, Store):
            raise TypeError("expected a Store")
        if not isinstance(ty, FuncType):
            raise TypeError("expected a FuncType")
        idx = FUNCTIONS.allocate((func, ty.results, access_caller))
        _func = ffi.wasmtime_func_t()
        ffi.wasmtime_func_new(
            store._context(),
            ty.ptr(),
            trampoline,
            idx,
            finalize,
            byref(_func))
        self._func = _func

    @classmethod
    def _from_raw(cls, func: ffi.wasmtime_func_t) -> "Func":
        ty: "Func" = cls.__new__(cls)
        ty._func = func
        return ty

    def type(self, store: Storelike) -> FuncType:
        """
        Gets the type of this func as a `FuncType`
        """
        ptr = ffi.wasmtime_func_type(store._context(), byref(self._func))
        return FuncType._from_ptr(ptr, None)

    def __call__(self, store: Storelike, *params: Any) -> Any:
        """
        Calls this function with the given parameters

        Parameters can either be a `Val` or a native python value which can be
        converted to a `Val` of the corresponding correct type

        Returns `None` if this func has 0 return types
        Returns a single value if the func has 1 return type
        Returns a list if the func has more than 1 return type

        Note that you can also use the `__call__` method and invoke a `Func` as
        if it were a function directly.
        """

        ty = self.type(store)
        param_tys = ty.params
        if len(params) > len(param_tys):
            raise WasmtimeError("too many parameters provided: given %s, expected %s" %
                                (len(params), len(param_tys)))
        if len(params) < len(param_tys):
            raise WasmtimeError("too few parameters provided: given %s, expected %s" %
                                (len(params), len(param_tys)))

        params_ptr = (ffi.wasmtime_val_t * len(params))()
        params_set = 0
        try:
            for val in params:
                params_ptr[params_set] = Val._convert_to_raw(store, param_tys[params_set], val)
                params_set += 1

            result_tys = ty.results
            results_ptr = (ffi.wasmtime_val_t * len(result_tys))()

            with enter_wasm(store) as trap:
                error = ffi.wasmtime_func_call(
                    store._context(),
                    byref(self._func),
                    params_ptr,
                    len(params),
                    results_ptr,
                    len(result_tys),
                    trap)
                if error:
                    raise WasmtimeError._from_ptr(error)
        finally:
            for i in range(0, params_set):
                ffi.wasmtime_val_unroot(byref(params_ptr[i]))

        results = []
        for i in range(0, len(result_tys)):
            results.append(Val._from_raw(store, results_ptr[i]).value)
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return results[0]
        else:
            return results

    def _as_extern(self) -> ffi.wasmtime_extern_t:
        union = ffi.wasmtime_extern_union(func=self._func)
        return ffi.wasmtime_extern_t(ffi.WASMTIME_EXTERN_FUNC, union)


class Caller:
    __ptr: "Optional[ctypes._Pointer[ffi.wasmtime_caller_t]]"
    __context: "Optional[ctypes._Pointer[ffi.wasmtime_context_t]]"

    def __init__(self, ptr: "ctypes._Pointer[ffi.wasmtime_caller_t]"):
        self.__ptr = ptr
        self.__context = ffi.wasmtime_caller_context(ptr)

    def __getitem__(self, name: str) -> AsExtern:
        """
        Looks up an export with `name` on the calling module.

        If `name` isn't defined on the calling module, or if the caller has gone
        away for some reason, then this will raise a `KeyError`. For more
        information about when this could fail see the `get` method which
        returns `None` on failure.
        """

        ret = self.get(name)
        if ret is None:
            raise KeyError("failed to find export {}".format(name))
        return ret

    def get(self, name: str) -> Optional[AsExtern]:
        """
        Looks up an export with `name` on the calling module.

        May return `None` if the export isn't found, if it's not a memory (for
        now), or if the caller has gone away and this `Caller` object has
        persisted too long.
        """

        # First convert to a raw name so we can typecheck our argument
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)

        # Next see if we've been invalidated
        if self.__ptr is None:
            return None

        # And if we're not invalidated we can perform the actual lookup
        item = ffi.wasmtime_extern_t()
        ok = ffi.wasmtime_caller_export_get(self.__ptr, name_buf, len(name_bytes), byref(item))
        if ok:
            return wrap_extern(item)
        else:
            return None

    def _context(self) -> "ctypes._Pointer[ffi.wasmtime_context_t]":
        if self.__context is None:
            raise ValueError("caller is no longer valid")
        return self.__context

    def _invalidate(self) -> None:
        self.__ptr = None
        self.__context = None


def extract_val(val: Val) -> Any:
    a = val.value
    if a is not None:
        return a
    return val


@ffi.wasmtime_func_callback_t
def trampoline(idx, caller, params, nparams, results, nresults):  # type: ignore
    caller = Caller(caller)
    try:
        func, result_tys, access_caller = FUNCTIONS.get(idx or 0)
        pyparams = []
        if access_caller:
            pyparams.append(caller)

        for i in range(0, nparams):
            pyparams.append(Val._from_raw(caller, params[i], owned=False).value)
        pyresults = func(*pyparams)
        if nresults == 0:
            if pyresults is not None:
                raise WasmtimeError(
                    "callback produced results when it shouldn't")
        elif nresults == 1:
            results[0] = Val._convert_to_raw(caller, result_tys[0], pyresults)
        else:
            if len(pyresults) != nresults:
                raise WasmtimeError("callback produced wrong number of results")
            for i, result in enumerate(pyresults):
                results[i] = Val._convert_to_raw(caller, result_tys[i], result)
        return 0
    except Exception as e:
        global LAST_EXCEPTION
        LAST_EXCEPTION = e
        trap = Trap("python exception")._consume()
        return cast(trap, c_void_p).value
    finally:
        caller._invalidate()


@CFUNCTYPE(None, c_void_p)
def finalize(idx):  # type: ignore
    FUNCTIONS.deallocate(idx or 0)



FUNCTIONS = Slab()


@contextmanager
def enter_wasm(store: Storelike):  # type: ignore
    try:
        trap = POINTER(ffi.wasm_trap_t)()
        yield byref(trap)
        if trap:
            trap_obj = Trap._from_ptr(trap)
            maybe_raise_last_exn()
            raise trap_obj
    except WasmtimeError:
        maybe_raise_last_exn()
        raise


def maybe_raise_last_exn() -> None:
    global LAST_EXCEPTION
    if LAST_EXCEPTION is None:
        return
    exn = LAST_EXCEPTION
    LAST_EXCEPTION = None
    raise exn
